#include <initguid.h>	/* Needed to define GUIDs */
#include <WinError.h>
#include <Strmif.h>		/* To declare IBaseFilter for DabRadcapOpenFromDirectShow */

// {7ADE9440-0C05-4fde-BCEF-A535B416536F}
DEFINE_GUID(PROPSETID_DAB, 
0x7ade9440, 0xc05, 0x4fde, 0xbc, 0xef, 0xa5, 0x35, 0xb4, 0x16, 0x53, 0x6f);

// {C93D7972-0F09-4feb-8823-486C6AA3CD08}
DEFINE_GUID(PROPSETID_DAB_W, 
0xc93d7972, 0xf09, 0x4feb, 0x88, 0x23, 0x48, 0x6c, 0x6a, 0xa3, 0xcd, 0x8);

typedef enum {
	PROPERTY_DAB_ID = 0,
	PROPERTY_DAB_SUBCHID,
	PROPERTY_DAB_ENSEMBLEID,
	PROPERTY_DAB_ENSEMBLEINFORMATION,
	PROPERTY_DAB_SERVICEINFORMATION,
	PROPERTY_DAB_SERVICESTATUS,
	PROPERTY_DAB_UNCORRECTEDERRORCOUNT,
	PROPERTY_DAB_DATETIME,
	PROPERTY_DAB_FORMAT,
	PROPERTY_DAB_CORRELATOR,
	PROPERTY_DAB_PAD
} PROPERTY_DAB;


#ifdef __cplusplus
extern "C" {
#else
typedef unsigned char bool;
#endif

	typedef enum { ServiceStatusPresent, ServiceStatusAbsent, ServiceStatusLost } ServiceStatus;
	typedef enum { Monaural, DiscreteStereo, ParametricStereo, Surround } SpatialEncoding;
	typedef enum { EncodingTypeMP2, EncodingTypeAAC } EncodingType;

	typedef struct {
		unsigned short Id;
		char Label[16];
		unsigned short LabelMask;
	} EnsembleIdentification;

	typedef struct {
		unsigned short Id;
		wchar_t Label[16];
		unsigned short LabelMask;
	} EnsembleIdentificationW;

	typedef struct {
		unsigned SCIdS;
		char ComponentLabel[12];	/* Note this should be [16], but changing it now would break too much existing code. Use ComponentInfoW and ServiceInfoW in new applications. */
		unsigned short ComponentLabelMask;
		bool Primary;
		bool ConditionalAccess;
		unsigned Applications;	/* Bit flags for each application type */
		unsigned TMId;
		union {
			struct {
				unsigned SubChId;
				unsigned ASCTy;
			} Audio;
			struct {
				unsigned SubChId;
				unsigned DSCTy;
			} StreamData;
			struct {
				unsigned FIDCId;
				unsigned DSCTy;
			} FIDC;
			struct {
				unsigned SubChId;
				unsigned DSCTy;
				unsigned SCId;
				bool SCCAFlag;
				bool DG;
				unsigned short PacketAddress;
				unsigned short SCCA;
			} PacketData;
		};
	} ComponentInfo;

	typedef struct {
		unsigned SCIdS;
		wchar_t ComponentLabel[16];
		unsigned short ComponentLabelMask;
		bool Primary;
		bool ConditionalAccess;
		unsigned Applications;	// Bit flags for each application type
		unsigned TMId;
		union {
			struct {
				unsigned SubChId;
				unsigned ASCTy;
			} Audio;
			struct {
				unsigned SubChId;
				unsigned DSCTy;
			} StreamData;
			struct {
				unsigned FIDCId;
				unsigned DSCTy;
			} FIDC;
			struct {
				unsigned SubChId;
				unsigned DSCTy;
				unsigned SCId;
				bool SCCAFlag;
				bool DG;
				unsigned short PacketAddress;
				unsigned short SCCA;
			} PacketData;
		};
	} ComponentInfoW;



	typedef struct {
		unsigned PD;				// 0=Program 1=Data
		unsigned SId;				// Service ID
		unsigned short ServiceLabelMask;
		char ServiceLabel[16];
		unsigned NumSC;
		ComponentInfo Component[12];
	} ServiceInfo;

	typedef struct {
		unsigned PD;				// 0=Program 1=Data
		unsigned SId;				// Service ID
		unsigned short ServiceLabelMask;
		wchar_t ServiceLabel[16];
		unsigned NumSC;
		ComponentInfoW Component[12];
	} ServiceInfoW;

	typedef struct {
		unsigned char FPad[2];
		unsigned char XPad[196];
	} PAD;

	typedef struct {
		FILETIME Utc;
		FILETIME Local;
	} DABDateTimeInfo;

	typedef struct {
		EncodingType Encoding;
		ULONG BitRate;
		ULONG SamplingRate;
		SpatialEncoding Spatial;
		BOOL SpectralBandReplication;
	} DABFormat;

	typedef struct {
		__int16 LatitudeOffset;
		__int16 LongitudeOffset;
		unsigned __int16 TimeDelay : 11;
		unsigned __int16 Valid : 1;
	} TIISubIdentifier;


	typedef struct {
		TIISubIdentifier SubId[24];
		__int16 LatitudeCourse;
		__int16 LongitudeCourse;
		UCHAR Valid;
		union {
			UCHAR Fine;
			struct {
				UCHAR LatitudeFine : 4;
				UCHAR LongitudeFine : 4;
			};
		};
	} TIIMainIdentifier;

	typedef struct {
		UCHAR Data[30];
		UCHAR CRC[2];
	} FIB;

	typedef FIB FIBBlock[3];

	/* Service-related functions */
	HANDLE __stdcall DabRadcapOpen (UINT waveInID);
	HANDLE __stdcall DabRadcapOpenFromDirectShow (IBaseFilter *pFilter);
	BOOL __stdcall DabRadcapGetEnsembleInformation (HANDLE hDev, EnsembleIdentification *pInfo);
	BOOL __stdcall DabRadcapGetEnsembleInformationW (HANDLE hDev, EnsembleIdentificationW *pInfo);
	BOOL __stdcall DabRadcapGetServiceInformation (HANDLE hDev, ServiceInfo *pInfo);
	BOOL __stdcall DabRadcapGetServiceInformationW (HANDLE hDev, ServiceInfoW *pInfo);
	BOOL __stdcall DabRadcapGetPAD (HANDLE hDev, PAD *pPad, OVERLAPPED *pOverlap);
	BOOL __stdcall DabRadcapGetServiceStatus (HANDLE hDev, ServiceStatus *pStatus);
	UINT __stdcall DabRadcapGetUncorrectedErrorCount (HANDLE hDev);
	BOOL __stdcall DabRadcapGetDateTime (HANDLE hDev, DABDateTimeInfo *pDateTime);
	BOOL __stdcall DabRadcapGetFormat (HANDLE hDev, DABFormat *pFormat);

	/* Card-related functions */
	UINT __stdcall DabRadcapGetNumberOfCards ();
	UINT __stdcall DabRadcapGetCardEnsembles (UINT CardNum);
	UINT __stdcall DabRadcapGetCardChannel (UINT CardNum, UINT EnsembleNum);
	BOOL __stdcall DabRadcapSetCardChannel (UINT CardNum, UINT EnsembleNum, UCHAR ChannelNum);
	BOOL __stdcall DabRadcapGetCardEnsembleInformation (UINT CardNum, UINT EnsembleNum, EnsembleIdentification *pInfo);
	BOOL __stdcall DabRadcapGetCardEnsembleInformationW (UINT CardNum, UINT EnsembleNum, EnsembleIdentificationW *pInfo);
	UINT __stdcall DabRadcapGetCardSignalStrength (UINT CardNum, UINT EnsembleNum);
	UINT __stdcall DabRadcapGetCardPhaseReferenceCorrelator (UINT CardNum, UINT EnsembleNum);
	BOOL __stdcall DabRadcapGetCardSpectrum (UINT CardNum, UINT EnsembleNum, float *SpectrumArray, SIZE_T SpectrumArraySize);
	BOOL __stdcall DabRadcapGetCardTII (UINT CardNum, UINT EnsembleNum, float *TIIArray, SIZE_T TIIArraySize);
	BOOL __stdcall DabRadcapGetCardTIIDatabaseEntry (UINT CardNum, UINT EnsembleNum, UINT MainIdentifier, TIIMainIdentifier *Entry);
	BOOL __stdcall DabRadcapGetCurrentFIBBlockIndex (UINT CardNum, UINT EnsembleNum, UCHAR *pFIBBlockIndex);
	BOOL __stdcall DabRadcapGetFIBBlock (UINT CardNum, UINT EnsembleNum, UCHAR Index, FIBBlock *pBlock);

#ifdef __cplusplus
}
#endif
