/******************************************************************************
** Monitor.cpp -- a simple DAB+ Radcap monitoring program                    **
*******************************************************************************
** Copyright (C) 2022 Sonifex Ltd                              **
**                                                                           **
** This software is provided 'as-is', without any express or implied         **
** warranty. In no event will the authors be held liable for any damages     **
** arising from the use of this software.                                    **
**                                                                           **
** Permission is granted to anyone to use this software for any purpose,     **
** including commercial applications, and to alter it and redistribute it    **
** freely, subject to the following restrictions:                            **
**   1. The origin of this software must not be misrepresented; you must not **
**      claim that you wrote the original software. If you use this software **
**      in a product, an acknowledgment in the product documentation would   **
**      be appreciated but is not required.                                  **
**   2. Altered source versions must be plainly marked as such, and must not **
**      be misrepresented as being the original software.                    **
**   3. This notice may not be removed or altered from any source            **
**      distribution.                                                        **
******************************************************************************/
#define STRICT
#include <windows.h>
#include <mmsystem.h>
#include <gdiplus.h>
#include <assert.h>
#include <process.h>
#include <strsafe.h>
#include "dabradcap.h"
#include "monitor.h"

using namespace Gdiplus;

struct EnsembleLocator {
	unsigned Card;
	unsigned Ensemble;
};

HINSTANCE Instance;
HWND hButton[ServiceMax];
HBRUSH hBackgroundBrush;
HBRUSH hBlackBrush;
HBRUSH hOffwhiteBrush;
HBRUSH hUnlitBrush;
HFONT hFont11;
HFONT hFont12;
HWND hwndMain;
HANDLE StopPad;
unsigned CaptureBuffers;
unsigned RenderBuffers;
UINT NumInputDevices;
UINT ServiceCount;
HWAVEIN hCapture;
HWAVEOUT hRender;
int OutstandingBuffers;
int ChannelSelection;
EnsembleLocator EnsembleLocation[MAX_ENSEMBLES];

WAVEFORMATEX Format = {
	WAVE_FORMAT_PCM,	// wFormatTag
	2,					// nChannels
	48000,				// nSamplesPerSec
	192000,				// nAvgBytesPerSec
	4,					// nBlockAlign
	16,					// wBitsPerSample
	0					// cbSize
};

int WINAPI WinMain (HINSTANCE hInstance, HINSTANCE, LPSTR, int cmdshow) {
	CoInitializeEx (NULL, COINIT_APARTMENTTHREADED);
	Instance = hInstance;
	hBackgroundBrush = CreateSolidBrush (RGB (216, 196, 144));
	hUnlitBrush = CreateSolidBrush (RGB (200, 199, 170));
	hBlackBrush = (HBRUSH) GetStockObject (BLACK_BRUSH);
	hOffwhiteBrush = CreateSolidBrush (RGB (255, 250, 228));

	WNDCLASS wc;
	wc.style = NULL;
	wc.lpfnWndProc = MonitorProc;
	wc.cbClsExtra = 0;
	wc.cbWndExtra = 0;
	wc.hInstance = hInstance;
	wc.hIcon = NULL;
	wc.hCursor = LoadCursor (NULL, IDC_ARROW);
	wc.hbrBackground = hBackgroundBrush;
	wc.lpszMenuName = NULL;
	wc.lpszClassName = L"MonitorClass";
	RegisterClass (&wc);

	wc.style = NULL;
	wc.lpfnWndProc = GenericButtonProc;
	wc.cbClsExtra = 0;
	wc.cbWndExtra = 0;
	wc.hInstance = hInstance;
	wc.hIcon = NULL;
	wc.hCursor = LoadCursor (NULL, IDC_ARROW);
	wc.hbrBackground = hBlackBrush;
	wc.lpszMenuName = NULL;
	wc.lpszClassName = L"GenericButtonClass";
	RegisterClass (&wc);

	wc.style = NULL;
	wc.lpfnWndProc = DlsBoxProc;
	wc.cbClsExtra = 0;
	wc.cbWndExtra = 0;
	wc.hInstance = hInstance;
	wc.hIcon = NULL;
	wc.hCursor = LoadCursor (NULL, IDC_ARROW);
	wc.hbrBackground = hOffwhiteBrush;
	wc.lpszMenuName = NULL;
	wc.lpszClassName = L"DlsBoxClass";
	RegisterClass (&wc);

	wc.style = NULL;
	wc.lpfnWndProc = InfoBoxProc;
	wc.cbClsExtra = 0;
	wc.cbWndExtra = 0;
	wc.hInstance = hInstance;
	wc.hIcon = NULL;
	wc.hCursor = LoadCursor (NULL, IDC_ARROW);
	wc.hbrBackground = hOffwhiteBrush;
	wc.lpszMenuName = NULL;
	wc.lpszClassName = L"InfoBoxClass";
	RegisterClass (&wc);

	wc.style = NULL;
	wc.lpfnWndProc = MotBoxProc;
	wc.cbClsExtra = 0;
	wc.cbWndExtra = 0;
	wc.hInstance = hInstance;
	wc.hIcon = NULL;
	wc.hCursor = LoadCursor (NULL, IDC_ARROW);
	wc.hbrBackground = hBlackBrush;
	wc.lpszMenuName = NULL;
	wc.lpszClassName = L"MotBoxClass";
	RegisterClass (&wc);

	wc.style = NULL;
	wc.lpfnWndProc = CardBoxProc;
	wc.cbClsExtra = 0;
	wc.cbWndExtra = 0;
	wc.hInstance = hInstance;
	wc.hIcon = NULL;
	wc.hCursor = LoadCursor (NULL, IDC_ARROW);
	wc.hbrBackground = hUnlitBrush;
	wc.lpszMenuName = NULL;
	wc.lpszClassName = L"CardBoxClass";
	RegisterClass (&wc);

	wc.style = NULL;
	wc.lpfnWndProc = ChannelButtonUnderlayProc;
	wc.cbClsExtra = 0;
	wc.cbWndExtra = 0;
	wc.hInstance = hInstance;
	wc.hIcon = NULL;
	wc.hCursor = LoadCursor (NULL, IDC_ARROW);
	wc.hbrBackground = hBackgroundBrush;
	wc.lpszMenuName = NULL;
	wc.lpszClassName = L"ChannelButtonUnderlayClass";
	RegisterClass (&wc);

	wc.style = NULL;
	wc.lpfnWndProc = RSSIIndicatorProc;
	wc.cbClsExtra = 0;
	wc.cbWndExtra = sizeof (void *);
	wc.hInstance = hInstance;
	wc.hIcon = NULL;
	wc.hCursor = LoadCursor (NULL, IDC_ARROW);
	wc.hbrBackground = hBlackBrush;
	wc.lpszMenuName = NULL;
	wc.lpszClassName = L"RSSIIndicatorClass";
	RegisterClass (&wc);

	wc.style = NULL;
	wc.lpfnWndProc = SpectrumProc;
	wc.cbClsExtra = 0;
	wc.cbWndExtra = 0;
	wc.hInstance = hInstance;
	wc.hIcon = NULL;
	wc.hCursor = LoadCursor (NULL, IDC_ARROW);
	wc.hbrBackground = (HBRUSH) GetStockObject (BLACK_BRUSH);
	wc.lpszMenuName = NULL;
	wc.lpszClassName = L"SpectrumClass";
	RegisterClass (&wc);

	LOGFONT Font = {
		-11,
		0,
		0,
		0,
		FW_NORMAL,
		FALSE,
		FALSE,
		FALSE,
		ANSI_CHARSET,
		OUT_DEFAULT_PRECIS,
		CLIP_DEFAULT_PRECIS,
		PROOF_QUALITY,
		DEFAULT_PITCH | FF_SWISS,
		L""
	};
	hFont11 = CreateFontIndirect (&Font);
	Font.lfHeight = -12;
	hFont12 = CreateFontIndirect (&Font);

	StopPad = CreateEvent (NULL, FALSE, FALSE, NULL);	// Used to stop the PAD receiver thread

   // Initialize GDI+.
	GdiplusStartupInput gdiplusStartupInput;
	ULONG_PTR gdiplusToken;
	GdiplusStartup(&gdiplusToken, &gdiplusStartupInput, NULL);

	// Centre the window on the screen
	RECT DesktopRect;
	SystemParametersInfo (SPI_GETWORKAREA, 0, &DesktopRect, 0);
	const int WindowWidth = 840;
	const int WindowHeight = 640;
	hwndMain = CreateWindowEx (0, L"MonitorClass", L"DAB+ Radcap Monitor", (WS_OVERLAPPEDWINDOW & ~WS_MAXIMIZEBOX & ~WS_THICKFRAME), (DesktopRect.left + DesktopRect.right - WindowWidth) / 2, (DesktopRect.top + DesktopRect.bottom - WindowHeight) / 2, WindowWidth, WindowHeight, NULL, NULL, hInstance, NULL);
	ShowWindow (hwndMain, cmdshow);
	SetPriorityClass (GetCurrentProcess (), REALTIME_PRIORITY_CLASS);
	SetThreadPriority (GetCurrentThread (), THREAD_PRIORITY_TIME_CRITICAL);
	MSG msg;
	BOOL rc;
	while ((rc = GetMessage (&msg, NULL, NULL, NULL)) != 0) {
		if (rc == -1) break;
		TranslateMessage (&msg);
		DispatchMessage (&msg);
	}
	CloseHandle (StopPad);
	GdiplusShutdown(gdiplusToken);
	DeleteObject (hBackgroundBrush);
	DeleteObject (hUnlitBrush);
	DeleteObject (hOffwhiteBrush);
	DeleteObject (hFont11);
	DeleteObject (hFont12);
	CoUninitialize ();
	return 0;
}


int _cdecl SIdComparator (const void *Element1, const void *Element2) {
	// To sort buttons into SId order
	return (int) GetWindowLongPtr (*(HWND *) Element1, GWLP_ID) - (int) GetWindowLongPtr (*(HWND *) Element2, GWLP_ID);
}


LRESULT CALLBACK MonitorProc (HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam) {
	switch (message) {
		case WM_CREATE:
			{
				NumInputDevices = waveInGetNumDevs ();
				RECT Rect;
				GetClientRect (hwnd, &Rect);
				RECT MotRect;
				MotRect.bottom = 240;
				MotRect.top = 0;
				MotRect.left = 0;
				MotRect.right = 320;
				AdjustWindowRectEx (&MotRect, WS_CHILD, FALSE, WS_EX_CLIENTEDGE);
				unsigned const WindowBottom = Rect.bottom;
				unsigned MotTop = WindowBottom - 6 + MotRect.top - MotRect.bottom;
				unsigned MotLeft = Rect.right - 6 + MotRect.left - MotRect.right;
				CreateWindowEx (WS_EX_CLIENTEDGE, L"MotBoxClass", NULL, WS_CHILD | WS_VISIBLE, MotLeft, MotTop, MotRect.right - MotRect.left, MotRect.bottom - MotRect.top, hwnd, (HMENU) MOT_BOX, Instance, NULL);
				CreateWindowEx (WS_EX_CLIENTEDGE, L"DlsBoxClass", NULL, WS_CHILD | WS_VISIBLE, 6, WindowBottom - 76, 190, 70, hwnd, (HMENU) DLS_BOX, Instance, NULL);
				CreateWindowEx (WS_EX_CLIENTEDGE, L"InfoBoxClass", NULL, WS_CHILD | WS_VISIBLE, 6, MotTop, 190, 160, hwnd, (HMENU) INFO_BOX, Instance, NULL);
				unsigned CardCount = DabRadcapGetNumberOfCards ();
				unsigned EnsembleNum = 0;
				for (unsigned i = 0; i < CardCount && EnsembleNum < MAX_ENSEMBLES; i++) {
					unsigned EnsembleCount = DabRadcapGetCardEnsembles (i);
					for (unsigned Ens = 0; Ens < EnsembleCount && EnsembleNum < MAX_ENSEMBLES; Ens++) {
						int CardPositionX = (Rect.left + Rect.right - CardCount * EnsembleCount * 204) / 2 + 204 * EnsembleNum;
						EnsembleLocation[EnsembleNum].Card = i;
						EnsembleLocation[EnsembleNum].Ensemble = Ens;
						CreateWindowEx (WS_EX_WINDOWEDGE, L"CardBoxClass", NULL, WS_CHILD | WS_VISIBLE | WS_DLGFRAME, CardPositionX, 5, 200, 80, hwnd, (HMENU)(0x20000 + EnsembleNum), Instance, NULL);
						++EnsembleNum;
					}
				}
				HWND ChannelButtonWindow = CreateWindowEx (WS_EX_WINDOWEDGE, L"ChannelButtonUnderlayClass", NULL, WS_CHILD | WS_VISIBLE | WS_DLGFRAME | WS_VSCROLL, 50, 110, Rect.right - 100 - SM_CXVSCROLL, 212, hwnd, (HMENU) CHANNEL_BUTTON_BOX, Instance, NULL);
				RECT ChannelButtonWindowRect;
				GetClientRect (ChannelButtonWindow, &ChannelButtonWindowRect);
				unsigned OffsetX = (ChannelButtonWindowRect.right - ChannelButtonWindowRect.left - 640) / 2;
				for (UINT WavDev = 0; ServiceCount < ServiceMax && WavDev < NumInputDevices; WavDev++) {
					// Determine which devices are DAB Radcap services
					HANDLE hDev = DabRadcapOpen (WavDev);
					if (hDev != INVALID_HANDLE_VALUE) {
						// It's a DAB service
						div_t RowColumn = div (ServiceCount, 8);
						// Obtain the short name for this service
						// Use only the characters in the service label indicated by
						// a corresponding 1-bit in the mask field
						ServiceInfoW Info;
						EnsembleIdentificationW EnsembleInfo;
						if (DabRadcapGetServiceInformationW (hDev, &Info) && DabRadcapGetEnsembleInformationW (hDev, &EnsembleInfo)) {
							WCHAR ServiceName[20];
							WCHAR *pName = ServiceName;
							unsigned short Mask = Info.ServiceLabelMask;
							for (unsigned m = 0; m < 16; m++, Mask <<= 1) {
								if (Mask & 0x8000) *(pName++) = Info.ServiceLabel[m];
							}
							*pName = 0;
							if (GetDlgItem (ChannelButtonWindow, Info.SId | (EnsembleInfo.Id << 16)) == NULL) {	// Make sure we don't have a duplicate SId
								div_t RowColumn = div (ServiceCount, 8);
								hButton[ServiceCount] = CreateWindowW (
									L"GenericButtonClass",
									ServiceName,
									WS_CHILD | WS_VISIBLE,
									OffsetX + RowColumn.rem * 80,
									10 + RowColumn.quot * 24,
									80,
									24,
									ChannelButtonWindow,
									(HMENU) (Info.SId | (EnsembleInfo.Id << 16)),
									Instance,
									(LPVOID) new ChannelButton (ChannelButtonWindow, Info.SId | (EnsembleInfo.Id << 16), hDev, WavDev)
								);
								++ServiceCount;
							}
							else CloseHandle (hDev);
						}
					}
				}
				// Sort the buttons by SId
				qsort (hButton, ServiceCount, sizeof (HWND), SIdComparator);
				unsigned Remainder = ServiceCount % 8;
				for (unsigned i = 0; i < ServiceCount - Remainder; i++) {
					div_t RowColumn = div (i, 8);
					MoveWindow (hButton[i], OffsetX + RowColumn.rem * 80, 10 + RowColumn.quot * 24, 80, 24, TRUE);
				}
				// Centre the last line of buttons in the window
				for (unsigned i = ServiceCount - Remainder; i < ServiceCount; i++) {
					div_t RowColumn = div (i, 8);
					MoveWindow (hButton[i], OffsetX + RowColumn.rem * 80 + (8 - Remainder) * 40, 10 + RowColumn.quot * 24, 80, 24, TRUE);
				}
				unsigned const ButtonHeight = 24;
				LONG ButtonArrayHeight = ((ServiceCount + 7) / 8) * ButtonHeight;
				SCROLLINFO siv = { 0 };
				siv.cbSize = sizeof (SCROLLINFO);
				siv.fMask = SIF_RANGE | SIF_PAGE;
				siv.nMin = 0;
				siv.nMax = ButtonArrayHeight + 20;	// Allow 10 pixwels above and below button array
				siv.nPage = ChannelButtonWindowRect.bottom - ChannelButtonWindowRect.top;
				SetScrollInfo (ChannelButtonWindow, SB_VERT, &siv, TRUE);
				SetTimer (hwnd, 1, 200, NULL);	// Used to check device status and update information display
			}
			return 0;
		case UM_BUTTONDOWN:
			{
				HWND hChannelButtonWindow = GetDlgItem (hwnd, CHANNEL_BUTTON_BOX);
				if (ChannelSelection) {
					SetDlgItemText (GetDlgItem (hwnd, INFO_BOX), DATE_TIME, L"");
					SetDlgItemText (GetDlgItem (hwnd, INFO_BOX), SERVICE_NAME, L"");
					SetDlgItemText (GetDlgItem (hwnd, INFO_BOX), SERVICE_ID, L"");
					SetDlgItemText (GetDlgItem (hwnd, INFO_BOX), SUBCHANNEL_ID, L"");
					SetDlgItemText (GetDlgItem (hwnd, INFO_BOX), ENSEMBLE_ID, L"");
					SetDlgItemText (GetDlgItem (hwnd, INFO_BOX), FORMAT, L"");
					SetDlgItemText (GetDlgItem (hwnd, INFO_BOX), SAMPLING_RATE, L"");
					SetDlgItemText (GetDlgItem (hwnd, INFO_BOX), BIT_RATE, L"");
					SetDlgItemText (GetDlgItem (hwnd, INFO_BOX), UNCORRECTED_ERRORS, L"");
					ChannelButton *pButton = (ChannelButton *) GetWindowLongPtr (GetDlgItem (hChannelButtonWindow, ChannelSelection), GWLP_USERDATA);
					SetEvent (StopPad);
					if (hCapture) {
						HWAVEIN hTemp = hCapture;
						hCapture = NULL;
						waveInReset (hTemp);
						// The device will be closed when the last buffer is returned
					}
					if (hRender) {
						HWAVEOUT hTemp = hRender;
						hRender = NULL;
						waveOutReset (hTemp);
						// The device will be closed when the last buffer is returned
					}
				}
				if (ChannelSelection == int (wParam)) {
					// Deselect the button
					ChannelSelection = 0;
				}
				else {
					// Deselect the old button (if any) and select the new one
					int OldSelection = ChannelSelection;
					ChannelSelection = int (wParam);
					if (OldSelection) {
						InvalidateRect (GetDlgItem (hChannelButtonWindow, OldSelection), NULL, FALSE);
					}
				}
				InvalidateRect (GetDlgItem (hChannelButtonWindow, int (wParam)), NULL, FALSE);
				if (ChannelSelection) PostMessage (hwnd, UM_OPENCHANNEL, 0, 0);	// Allow the MIM_DATA and WOM_DONE messages to be processed first
			}
			return 0;
		case UM_OPENCHANNEL:
			if (ChannelSelection) {
				HWND hChannelButtonWindow = GetDlgItem (hwnd, CHANNEL_BUTTON_BOX);
				OutstandingBuffers = 0;
				ChannelButton *pButton = (ChannelButton *) GetWindowLongPtr (GetDlgItem (hChannelButtonWindow, ChannelSelection), GWLP_USERDATA);
				if (pButton) {
					if (waveOutOpen (&hRender, 0, &Format, (DWORD_PTR) hwnd, 0, CALLBACK_WINDOW) == MMSYSERR_NOERROR) {
						waveOutPause (hRender);
					}
					else hRender = NULL;
					if (waveInOpen (&hCapture, pButton->GetWaveId (), &Format, (DWORD_PTR) hwnd, NULL, CALLBACK_WINDOW) == MMSYSERR_NOERROR) {
						// Create 4 buffers and pass them to the capture device
						for (unsigned i = 0; i < 4; i++) {
							WAVEHDR *pHeader = new WAVEHDR;
							if (pHeader) {
								pHeader->dwBufferLength = Format.nAvgBytesPerSec * 3 / 25;	// 120ms
								pHeader->dwBytesRecorded = 0;
								pHeader->dwFlags = 0;
								pHeader->dwLoops = 0;
								pHeader->dwUser = 0;
								pHeader->lpData = (LPSTR) new UCHAR[pHeader->dwBufferLength];
								waveInPrepareHeader (hCapture, pHeader, sizeof (WAVEHDR));
								waveInAddBuffer (hCapture, pHeader, sizeof (WAVEHDR));
								++CaptureBuffers;
							}
						}
						waveInStart (hCapture);
						pButton->ActivatePad ();
					}
				}
				else {
					ChannelSelection = 0;
				}
			}
			return 0;

		case WM_TIMER:
			{
				HWND ChannelButtonWindow = GetDlgItem (hwnd, CHANNEL_BUTTON_BOX);
				RECT ChannelButtonWindowRect;
				GetClientRect (ChannelButtonWindow, &ChannelButtonWindowRect);
				unsigned OffsetX = (ChannelButtonWindowRect.right + ChannelButtonWindowRect.left - 640) / 2;
				for (unsigned i = 0; i < ServiceCount; i++) {
					ChannelButton *pButton = (ChannelButton *) GetWindowLongPtr (hButton[i], GWLP_USERDATA);
					ServiceStatus Status = ServiceStatusLost;
					if (pButton && pButton->GetServiceStatus (&Status) == FALSE || Status == ServiceStatusLost) {
						if (int (GetWindowLongPtr (hButton[i], GWLP_ID)) == ChannelSelection) {
							SetDlgItemText (GetDlgItem (hwnd, INFO_BOX), DATE_TIME, L"");
							SetDlgItemText (GetDlgItem (hwnd, INFO_BOX), SERVICE_NAME, L"");
							SetDlgItemText (GetDlgItem (hwnd, INFO_BOX), SERVICE_ID, L"");
							SetDlgItemText (GetDlgItem (hwnd, INFO_BOX), SUBCHANNEL_ID, L"");
							SetDlgItemText (GetDlgItem (hwnd, INFO_BOX), ENSEMBLE_ID, L"");
							SetDlgItemText (GetDlgItem (hwnd, INFO_BOX), FORMAT, L"");
							SetDlgItemText (GetDlgItem (hwnd, INFO_BOX), SAMPLING_RATE, L"");
							SetDlgItemText (GetDlgItem (hwnd, INFO_BOX), BIT_RATE, L"");
							SetDlgItemText (GetDlgItem (hwnd, INFO_BOX), UNCORRECTED_ERRORS, L"");
							SetEvent (StopPad);
							if (hCapture) {
								HWAVEIN hTemp = hCapture;
								hCapture = NULL;
								waveInReset (hTemp);
							}
							if (hRender) {
								HWAVEOUT hTemp = hRender;
								hRender = NULL;
								waveOutReset (hTemp);
							}
							ChannelSelection = 0;
						}
						DestroyWindow (hButton[i]);
						--ServiceCount;
						for (unsigned j = i; j < ServiceCount; j++) {
							div_t RowColumn = div (j, 8);
							hButton[j] = hButton[j + 1];
							MoveWindow (
								hButton[j],
								OffsetX + RowColumn.rem * 80,
								10 + RowColumn.quot * 24,
								80,
								24,
								TRUE );
						}
						// Centre the last line of buttons in the window
						unsigned Remainder = ServiceCount % 8;
						for (unsigned k = ServiceCount - Remainder; k < ServiceCount; k++) {
							div_t RowColumn = div (k, 8);
							MoveWindow (hButton[k], OffsetX + RowColumn.rem * 80 + (8 - Remainder) * 40, 10 + RowColumn.quot * 24, 80, 24, TRUE);
						}
						--i;
						unsigned const ButtonHeight = 24;
						LONG ButtonArrayHeight = ((ServiceCount + 7) / 8) * ButtonHeight;
						SCROLLINFO siv = { 0 };
						siv.cbSize = sizeof (SCROLLINFO);
						siv.fMask = SIF_RANGE | SIF_PAGE;
						siv.nMin = 0;
						siv.nMax = ButtonArrayHeight + 20;	// Allow 10 pixwels above and below button array
						siv.nPage = ChannelButtonWindowRect.bottom - ChannelButtonWindowRect.top;
						SetScrollInfo (ChannelButtonWindow, SB_VERT, &siv, TRUE);
					}
					else {
						if (int (GetWindowLongPtr (hButton[i], GWLP_ID)) == ChannelSelection) {
							// Service is still valid
							WCHAR String[40];
							if (Status == ServiceStatusAbsent) {
								SetDlgItemText (GetDlgItem (hwnd, INFO_BOX), DATE_TIME, L"--- No Signal ---");
							}
							else {
								DABDateTimeInfo DateTime;
								if (pButton->GetDateTime (&DateTime)) {
									SYSTEMTIME LocalTime;
									FileTimeToSystemTime (&DateTime.Local, &LocalTime);
									StringCbPrintfW (String, sizeof String, L"%2d %s %4.4d %2.2d:%2.2d:%2.2d", LocalTime.wDay, MonthName[LocalTime.wMonth - 1], LocalTime.wYear, LocalTime.wHour, LocalTime.wMinute, LocalTime.wSecond);
									SetDlgItemTextW (GetDlgItem (hwnd, INFO_BOX), DATE_TIME, String);
								}
							}
							EnsembleIdentificationW EnsembleInfo;
							if (pButton->GetEnsembleInfoW (&EnsembleInfo)) {
								StringCbPrintfW (String, sizeof String, L"Ensemble ID : %4.4x", EnsembleInfo.Id);
								SetDlgItemTextW (GetDlgItem (hwnd, INFO_BOX), ENSEMBLE_ID, String);
							}
							ServiceInfoW Info;
							if (pButton->GetServiceInfoW (&Info)) {
								// Info.ServiceLabel isn't NULL terminated, so copy it into our String buffer and terminate it
								memcpy (String, Info.ServiceLabel, sizeof Info.ServiceLabel);
								String[16] = 0;
								// Remove trailing spaces
								for (int i = 15; i >= 0 && String[i] == L' '; i--) String[i] = 0;
								SetDlgItemTextW (GetDlgItem (hwnd, INFO_BOX), SERVICE_NAME, String);
								StringCbPrintfW (String, sizeof String, L"SID : %4.4x  SCIdS : %d", Info.SId, Info.Component[0].SCIdS);
								SetDlgItemTextW (GetDlgItem (hwnd, INFO_BOX), SERVICE_ID, String);
								StringCbPrintfW (String, sizeof String, L"SubChID : %d", Info.Component[0].Audio.SubChId);
								SetDlgItemTextW (GetDlgItem (hwnd, INFO_BOX), SUBCHANNEL_ID, String);
							}
							DABFormat Format;
							if (pButton->GetFormat (&Format)) {
								StringCbPrintfW (String, sizeof String, L"%d samples / sec", Format.SamplingRate);
								SetDlgItemTextW (GetDlgItem (hwnd, INFO_BOX), SAMPLING_RATE, String);
								StringCbPrintfW (String, sizeof String, L"%d kbps", Format.BitRate / 1000);
								SetDlgItemTextW (GetDlgItem (hwnd, INFO_BOX), BIT_RATE, String);
								SetDlgItemTextW (GetDlgItem (hwnd, INFO_BOX), FORMAT,
									Format.Encoding == EncodingTypeMP2 ? FormatStringMP2[Format.Spatial] :
									Format.SpectralBandReplication ? FormatStringAACSBR[Format.Spatial] : FormatStringAAC[Format.Spatial]);
							}
							StringCbPrintfW (String, sizeof String, L"%d uncorrected errors", pButton->GetUncorrectedErrorCount ());				
							SetDlgItemTextW (GetDlgItem (hwnd, INFO_BOX), UNCORRECTED_ERRORS, String);
						}
					}
				}
				UINT InputDeviceCount = waveInGetNumDevs ();
				UINT FoundServices = 0;
				for (UINT WavDev = 0; WavDev < InputDeviceCount; ++WavDev) {
					HANDLE hDev = DabRadcapOpen (WavDev);
					if (hDev != INVALID_HANDLE_VALUE) {
						ServiceStatus Status = ServiceStatusLost;
						if (DabRadcapGetServiceStatus (hDev, &Status) && Status != ServiceStatusLost) {
							ServiceInfoW Info;
							DabRadcapGetServiceInformationW (hDev, &Info);
							EnsembleIdentificationW EnsembleInfo;
							DabRadcapGetEnsembleInformationW (hDev, &EnsembleInfo);
							WCHAR ServiceName[20];
							WCHAR *pName = ServiceName;
							unsigned short Mask = Info.ServiceLabelMask;
							for (unsigned m = 0; m < 16; m++, Mask <<= 1) {
								if (Mask & 0x8000) *(pName++) = Info.ServiceLabel[m];
							}
							*pName = 0;
							HWND ThisButton = GetDlgItem (ChannelButtonWindow, Info.SId | (EnsembleInfo.Id << 16));
							if (ThisButton == NULL) {
								// It's a new DAB service
								if (ServiceCount < ServiceMax) {
									div_t RowColumn = div (ServiceCount, 8);
									// Obtain the short name for this service
									// Use only the characters in the service label indicated by
									// a corresponding 1-bit in the mask field
									hButton[ServiceCount] = CreateWindowW (
										L"GenericButtonClass",
										ServiceName,
										WS_CHILD | WS_VISIBLE,
										OffsetX + RowColumn.rem * 80,
										100 + RowColumn.quot * 24,
										80,
										24,
										ChannelButtonWindow,
										(HMENU) (Info.SId | (EnsembleInfo.Id << 16)),
										Instance,
										(LPVOID) new ChannelButton (ChannelButtonWindow, Info.SId | (EnsembleInfo.Id << 16), hDev, WavDev)
									);
									++ServiceCount;
									// Sort the buttons by SId
									qsort (hButton, ServiceCount, sizeof (HWND), SIdComparator);
									// Centre the last line of buttons in the window
									unsigned Remainder = ServiceCount % 8;
									for (unsigned i = 0; i < ServiceCount - Remainder; i++) {
										div_t RowColumn = div (i, 8);
										MoveWindow (hButton[i], OffsetX + RowColumn.rem * 80, 10 + RowColumn.quot * 24, 80, 24, TRUE);
									}
									for (unsigned i = ServiceCount - Remainder; i < ServiceCount; i++) {
										div_t RowColumn = div (i, 8);
										MoveWindow (hButton[i], OffsetX + RowColumn.rem * 80 + (8 - Remainder) * 40, 10 + RowColumn.quot * 24, 80, 24, TRUE);
									}
									InvalidateRect (ChannelButtonWindow, NULL, FALSE);
									unsigned const ButtonHeight = 24;
									LONG ButtonArrayHeight = ((ServiceCount + 7) / 8) * ButtonHeight;
									SCROLLINFO siv = { 0 };
									siv.cbSize = sizeof (SCROLLINFO);
									siv.fMask = SIF_RANGE | SIF_PAGE;
									siv.nMin = 0;
									siv.nMax = ButtonArrayHeight + 20;	// Allow 10 pixwels above and below button array
									siv.nPage = ChannelButtonWindowRect.bottom - ChannelButtonWindowRect.top;
									SetScrollInfo (ChannelButtonWindow, SB_VERT, &siv, TRUE);
								}
								else {
									CloseHandle (hDev);
								}
							}
							else {
								// Button exists, so check to see if the service name has changed
								ChannelButton *pButton = (ChannelButton *) GetWindowLongPtr (ThisButton, GWLP_USERDATA);
								if (pButton) {
									pButton->SetWaveId (WavDev);
								}
								WCHAR ButtonName[20];
								GetWindowTextW (ThisButton, ButtonName, sizeof ButtonName);
								if (_wcsicmp (ServiceName, ButtonName) != 0) SetWindowTextW (ThisButton, ServiceName);
								CloseHandle (hDev);
							}
						}
						else CloseHandle (hDev);
					}
				}
				NumInputDevices = InputDeviceCount;
			}
			return 0;

		case MM_WIM_DATA:
			{
				// Returned capture buffers are normally passed to the render device
				HWAVEIN hWaveIn = (HWAVEIN) wParam;
				WAVEHDR *pHeader = (WAVEHDR *) lParam;
				if (hWaveIn && pHeader) {
					--CaptureBuffers;
					waveInUnprepareHeader (hWaveIn, pHeader, sizeof (WAVEHDR));
					if (hCapture) {
						if ((pHeader->dwFlags & WHDR_DONE) && pHeader->dwBytesRecorded > 0) {
							if (hRender && OutstandingBuffers < 3) {
								pHeader->dwFlags = 0;
								waveOutPrepareHeader (hRender, pHeader, sizeof (WAVEHDR));
								waveOutWrite (hRender, pHeader, sizeof (WAVEHDR));
								++RenderBuffers;
								// Restart playing once there are at least two outstanding buffers
								if (++OutstandingBuffers > 1) waveOutRestart (hRender);
							}
							else {
								// Capture is getting ahead of render so just skip this buffer
								waveInPrepareHeader (hCapture, pHeader, sizeof (WAVEHDR));
								waveInAddBuffer (hCapture, pHeader, sizeof (WAVEHDR));
								++CaptureBuffers;
							}
						}
						else {
							if (pHeader->lpData) delete[] pHeader->lpData;
							delete pHeader;
						}
					}
					else {
						if (pHeader->lpData) delete[] pHeader->lpData;
						delete pHeader;
						if (CaptureBuffers == 0) {
							waveInClose (hWaveIn);
						}
					}
				}
			}
			return 0;
		case MM_WOM_DONE:
			{
				// Returned render buffers are normally passed to the capture device
				HWAVEOUT hWaveOut = (HWAVEOUT) wParam;
				WAVEHDR *pHeader = (WAVEHDR *) lParam;
				if (hWaveOut && pHeader) {
					--RenderBuffers;
					waveOutUnprepareHeader (hWaveOut, pHeader, sizeof (WAVEHDR));
					if (hCapture) {
						pHeader->dwFlags = 0;
						waveInPrepareHeader (hCapture, pHeader, sizeof (WAVEHDR));
						waveInAddBuffer (hCapture, pHeader, sizeof (WAVEHDR));
						++CaptureBuffers;
					}
					else {
						if (pHeader->lpData) delete[] pHeader->lpData;
						delete pHeader;
						if (RenderBuffers == 0) {
							waveOutClose (hWaveOut);
						}
					}
					// If capture is falling behind render, wait for it to catch up
					if (--OutstandingBuffers == 0 && hRender) waveOutPause (hRender);
				}
			}
			return 0;
		case WM_CLOSE:
			KillTimer (hwnd, 1);
			SetEvent (StopPad);
			if (hCapture) {
				HWAVEIN hTemp = hCapture;
				hCapture = NULL;
				SetEvent (StopPad);
				waveInReset (hTemp);
			}
			if (hRender) {
				HWAVEOUT hTemp = hRender;
				hRender = NULL;
				waveOutReset (hTemp);
			}
			DestroyWindow (hwnd);
			return 0;
		case WM_DESTROY:
			PostQuitMessage (0);
			return 0;
		default:
			return DefWindowProc (hwnd, message, wParam, lParam);
	}
}


LRESULT CALLBACK ChannelButtonUnderlayProc (HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam) {
	switch (message) {
		case WM_CREATE:
			{
			}
			return 0;
		case WM_VSCROLL:
			{
				SCROLLINFO si = { 0 };
				si.cbSize = sizeof (SCROLLINFO);
				si.fMask = SIF_ALL;
				GetScrollInfo (hwnd, SB_VERT, &si);
				int OldPosition = si.nPos;
				si.fMask = SIF_POS;
				switch (LOWORD (wParam)) {
				case SB_TOP:
					si.nPos = si.nMin;
					break;
              
				case SB_BOTTOM:
					si.nPos = si.nMax;
					break;
              
				case SB_LINEUP:
					si.nPos -= 1;
					break;

				case SB_LINEDOWN:
					si.nPos += 1;
					break;

				case SB_PAGEUP:
					si.nPos -= 20;
					break;

				case SB_PAGEDOWN:
					si.nPos += 20;
					break;

				case SB_THUMBTRACK:
					si.nPos = si.nTrackPos;
					break;
				default:
					return DefWindowProc (hwnd, message, wParam, lParam);
				}
				SetScrollInfo (hwnd, SB_VERT, &si, TRUE);
				GetScrollInfo (hwnd, SB_VERT, &si);
				if (OldPosition != si.nPos) {
					ScrollWindow (hwnd, 0, OldPosition - si.nPos, NULL, NULL);
					UpdateWindow (hwnd);
				}
			}
			return 0;
		case UM_BUTTONDOWN:
			SendMessage (GetParent (hwnd), UM_BUTTONDOWN, wParam, lParam);	// Pass this down to the underlying window
			return 0;
	}
	return DefWindowProc (hwnd, message, wParam, lParam);
}


LRESULT CALLBACK CardBoxProc (HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam) {
	switch (message) {
		case WM_CREATE:
			{
				RECT ClientRect;
				GetClientRect (hwnd, &ClientRect);
				CreateWindowEx (0, L"combobox", NULL, WS_CHILD | WS_VISIBLE | CBS_DROPDOWNLIST | WS_VSCROLL, ClientRect.right - 46, 2, 44, ClientRect.bottom - 2, hwnd, (HMENU) CHANNEL, Instance, NULL);
				SendDlgItemMessage (hwnd, CHANNEL, WM_SETFONT, (WPARAM) hFont11, 0);
				CreateWindowEx (0, L"static", NULL, WS_CHILD | WS_VISIBLE | SS_LEFT, ClientRect.right - 46, 30, 44, 15, hwnd, (HMENU) ENSEMBLE_ID, Instance, NULL);
				SendDlgItemMessage (hwnd, ENSEMBLE_ID, WM_SETFONT, (WPARAM) hFont11, 0);
				CreateWindowEx (0, L"static", NULL, WS_CHILD | WS_VISIBLE | SS_CENTER | SS_NOPREFIX, 2, 2, 120, 15, hwnd, (HMENU) ENSEMBLE_NAME, Instance, NULL);
				SendDlgItemMessage (hwnd, ENSEMBLE_NAME, WM_SETFONT, (WPARAM) hFont11, 0);
				CreateWindowEx (WS_EX_CLIENTEDGE, L"RSSIIndicatorClass", NULL, WS_CHILD | WS_VISIBLE, 125, 2, 16, ClientRect.bottom - 4, hwnd, (HMENU) CORRELATOR, Instance, NULL);
				CreateWindowEx (WS_EX_CLIENTEDGE, L"SpectrumClass", NULL, WS_CHILD | WS_VISIBLE, 2, 18, 120, ClientRect.bottom - 20, hwnd, (HMENU) SPECTRUM, Instance, NULL);
				SendDlgItemMessage (hwnd, CHANNEL, CB_ADDSTRING, 0, (LPARAM) L"5A");
				SendDlgItemMessage (hwnd, CHANNEL, CB_ADDSTRING, 0, (LPARAM) L"5B");
				SendDlgItemMessage (hwnd, CHANNEL, CB_ADDSTRING, 0, (LPARAM) L"5C");
				SendDlgItemMessage (hwnd, CHANNEL, CB_ADDSTRING, 0, (LPARAM) L"5D");
				SendDlgItemMessage (hwnd, CHANNEL, CB_ADDSTRING, 0, (LPARAM) L"6A");
				SendDlgItemMessage (hwnd, CHANNEL, CB_ADDSTRING, 0, (LPARAM) L"6B");
				SendDlgItemMessage (hwnd, CHANNEL, CB_ADDSTRING, 0, (LPARAM) L"6C");
				SendDlgItemMessage (hwnd, CHANNEL, CB_ADDSTRING, 0, (LPARAM) L"6D");
				SendDlgItemMessage (hwnd, CHANNEL, CB_ADDSTRING, 0, (LPARAM) L"7A");
				SendDlgItemMessage (hwnd, CHANNEL, CB_ADDSTRING, 0, (LPARAM) L"7B");
				SendDlgItemMessage (hwnd, CHANNEL, CB_ADDSTRING, 0, (LPARAM) L"7C");
				SendDlgItemMessage (hwnd, CHANNEL, CB_ADDSTRING, 0, (LPARAM) L"7D");
				SendDlgItemMessage (hwnd, CHANNEL, CB_ADDSTRING, 0, (LPARAM) L"8A");
				SendDlgItemMessage (hwnd, CHANNEL, CB_ADDSTRING, 0, (LPARAM) L"8B");
				SendDlgItemMessage (hwnd, CHANNEL, CB_ADDSTRING, 0, (LPARAM) L"8C");
				SendDlgItemMessage (hwnd, CHANNEL, CB_ADDSTRING, 0, (LPARAM) L"8D");
				SendDlgItemMessage (hwnd, CHANNEL, CB_ADDSTRING, 0, (LPARAM) L"9A");
				SendDlgItemMessage (hwnd, CHANNEL, CB_ADDSTRING, 0, (LPARAM) L"9B");
				SendDlgItemMessage (hwnd, CHANNEL, CB_ADDSTRING, 0, (LPARAM) L"9C");
				SendDlgItemMessage (hwnd, CHANNEL, CB_ADDSTRING, 0, (LPARAM) L"9D");
				SendDlgItemMessage (hwnd, CHANNEL, CB_ADDSTRING, 0, (LPARAM) L"10A");
				SendDlgItemMessage (hwnd, CHANNEL, CB_ADDSTRING, 0, (LPARAM) L"10N");
				SendDlgItemMessage (hwnd, CHANNEL, CB_ADDSTRING, 0, (LPARAM) L"10B");
				SendDlgItemMessage (hwnd, CHANNEL, CB_ADDSTRING, 0, (LPARAM) L"10C");
				SendDlgItemMessage (hwnd, CHANNEL, CB_ADDSTRING, 0, (LPARAM) L"10D");
				SendDlgItemMessage (hwnd, CHANNEL, CB_ADDSTRING, 0, (LPARAM) L"11A");
				SendDlgItemMessage (hwnd, CHANNEL, CB_ADDSTRING, 0, (LPARAM) L"11N");
				SendDlgItemMessage (hwnd, CHANNEL, CB_ADDSTRING, 0, (LPARAM) L"11B");
				SendDlgItemMessage (hwnd, CHANNEL, CB_ADDSTRING, 0, (LPARAM) L"11C");
				SendDlgItemMessage (hwnd, CHANNEL, CB_ADDSTRING, 0, (LPARAM) L"11D");
				SendDlgItemMessage (hwnd, CHANNEL, CB_ADDSTRING, 0, (LPARAM) L"12A");
				SendDlgItemMessage (hwnd, CHANNEL, CB_ADDSTRING, 0, (LPARAM) L"12N");
				SendDlgItemMessage (hwnd, CHANNEL, CB_ADDSTRING, 0, (LPARAM) L"12B");
				SendDlgItemMessage (hwnd, CHANNEL, CB_ADDSTRING, 0, (LPARAM) L"12C");
				SendDlgItemMessage (hwnd, CHANNEL, CB_ADDSTRING, 0, (LPARAM) L"12D");
				SendDlgItemMessage (hwnd, CHANNEL, CB_ADDSTRING, 0, (LPARAM) L"13A");
				SendDlgItemMessage (hwnd, CHANNEL, CB_ADDSTRING, 0, (LPARAM) L"13B");
				SendDlgItemMessage (hwnd, CHANNEL, CB_ADDSTRING, 0, (LPARAM) L"13C");
				SendDlgItemMessage (hwnd, CHANNEL, CB_ADDSTRING, 0, (LPARAM) L"13D");
				SendDlgItemMessage (hwnd, CHANNEL, CB_ADDSTRING, 0, (LPARAM) L"13E");
				SendDlgItemMessage (hwnd, CHANNEL, CB_ADDSTRING, 0, (LPARAM) L"13F");
				unsigned EnsembleNum = (unsigned) GetWindowLongPtr (hwnd, GWLP_ID) - 0x20000;
				UINT Channel = DabRadcapGetCardChannel (EnsembleLocation[EnsembleNum].Card, EnsembleLocation[EnsembleNum].Ensemble);
				SetWindowLongPtr (hwnd, GWLP_USERDATA, Channel);	// For checking for external channel change
				SendDlgItemMessage (hwnd, CHANNEL, CB_SETCURSEL, Channel, 0);
				SetTimer (hwnd, 1, 500, NULL);
			}
			return 0;
		case WM_COMMAND:
			switch (LOWORD(wParam)) {
				case CHANNEL:
					if (HIWORD (wParam) == CBN_SELENDOK) {
						unsigned Channel = (unsigned) SendDlgItemMessage (hwnd, CHANNEL, CB_GETCURSEL, 0, 0);
						unsigned EnsembleNum = (unsigned) GetWindowLongPtr (hwnd, GWLP_ID) - 0x20000;
						DabRadcapSetCardChannel (EnsembleLocation[EnsembleNum].Card, EnsembleLocation[EnsembleNum].Ensemble, Channel);
						return 0;
					}
					break;
    		}
			break;
		case WM_CTLCOLORSTATIC:
			SetTextColor ((HDC) wParam, RGB (0, 0, 0));
			SetBkColor ((HDC) wParam, RGB (200, 199, 170));
			return (LRESULT) hUnlitBrush;

		case WM_TIMER:
			if (wParam == 1) {
				WCHAR IdString[9] = L"ID: ";
				WCHAR EnsembleName[17] = L"";
				EnsembleIdentificationW Info;
				unsigned EnsembleNum = (unsigned) GetWindowLongPtr (hwnd, GWLP_ID) - 0x20000;
				// Make sure nothing else has changed channel!
				UINT Channel = DabRadcapGetCardChannel (EnsembleLocation[EnsembleNum].Card, EnsembleLocation[EnsembleNum].Ensemble);
				if (GetWindowLongPtr (hwnd, GWLP_USERDATA) != Channel) {
					SendDlgItemMessage (hwnd, CHANNEL, CB_SETCURSEL, Channel, 0);
					SetWindowLong (hwnd, GWLP_USERDATA, Channel);
				}
				if (DabRadcapGetCardEnsembleInformationW (EnsembleLocation[EnsembleNum].Card, EnsembleLocation[EnsembleNum].Ensemble, &Info)) {
					StringCbPrintf (IdString + 4, sizeof IdString - 4, L"%4.4X", Info.Id);
					wcsncpy_s (EnsembleName, sizeof EnsembleName / sizeof EnsembleName[0], Info.Label, 16);
					// Remove any trailing spaces
					for (int i = 15; i >= 0; i--) {
						if (EnsembleName[i] == ' ') EnsembleName[i] = 0;
						else break;
					}
				}
				SetDlgItemTextW (hwnd, ENSEMBLE_NAME, EnsembleName);
				SetDlgItemText (hwnd, ENSEMBLE_ID, IdString);
				unsigned Correlator = DabRadcapGetCardPhaseReferenceCorrelator (EnsembleLocation[EnsembleNum].Card, EnsembleLocation[EnsembleNum].Ensemble);
				SetWindowLongPtr (GetDlgItem (hwnd, CORRELATOR), GWLP_USERDATA, Correlator);
				InvalidateRect (GetDlgItem (hwnd, CORRELATOR), NULL, FALSE);
				POINT *pData = (POINT *) GetWindowLongPtr (GetDlgItem (hwnd, SPECTRUM), GWLP_USERDATA);
				if (pData) {
					float SpectrumBuffer[2048];
					if (DabRadcapGetCardSpectrum (EnsembleLocation[EnsembleNum].Card, EnsembleLocation[EnsembleNum].Ensemble, SpectrumBuffer, sizeof SpectrumBuffer)) {
						for (int i = 0; i < 2048; i++) {
							pData[i].x = i;
							pData[i].y = 600 - long (10 * SpectrumBuffer[i]);
						}
						pData[2048].x = 2048;
						pData[2048].y = 600;
					}
					else {
						for (int i = 0; i < 2049; i++) {
							pData[i].x = i;
							pData[i].y = 600;
						}
					}
					InvalidateRect (GetDlgItem (hwnd, SPECTRUM), NULL, FALSE);
				}
				return 0;
			}
			break;
		case WM_CLOSE:
			KillTimer (hwnd, 1);
	}
	return DefWindowProc (hwnd, message, wParam, lParam);
}


LRESULT CALLBACK RSSIIndicatorProc (HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam) {
	switch (message) {
		case WM_CREATE:
			SetWindowLongPtr (hwnd, 0, (LONG_PTR) new RSSIImage (hwnd));
			return 0;

		case WM_DESTROY:
			{
				RSSIImage *pImage = (RSSIImage *) GetWindowLongPtr (hwnd, 0);
				if (pImage) delete pImage;
			}
			break;

		case WM_PAINT:
			{
				PAINTSTRUCT PaintStruct;
				HDC hdc = BeginPaint (hwnd, &PaintStruct);
				RSSIImage *pImage = (RSSIImage *) GetWindowLongPtr (hwnd, 0);
				if (pImage) {
					RECT Rect;
					GetClientRect (hwnd, &Rect);
					int Reading = (int) GetWindowLongPtr (hwnd, GWLP_USERDATA);
					int Peak = Reading * Rect.bottom / 100;
					if (Peak > 0) {
						BitBlt (hdc, Rect.left, Rect.bottom - Peak, Rect.right, Peak, pImage->hMemoryDC, Rect.left, Rect.bottom - Peak, SRCCOPY);
					}
					Rect.bottom -= Peak;
					if (Rect.bottom > Rect.top) {
						FillRect (hdc, &Rect, (HBRUSH) GetStockObject (BLACK_BRUSH));
					}
				}
				EndPaint (hwnd, &PaintStruct);
			}
			return 0;

	}
	return DefWindowProc (hwnd, message, wParam, lParam);
}


LRESULT CALLBACK SpectrumProc (HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam) {
	switch (message) {
		case WM_CREATE:
			{
				LPCREATESTRUCT pCreate = (LPCREATESTRUCT) lParam;
				POINT *pData = new POINT[2049];
				for (int i = 0; i < 2049; i++) {
					pData[i].x = i;
					pData[i].y = 600;
				}
				SetWindowLongPtr (hwnd, GWLP_USERDATA, (LONG_PTR) pData);
			}
			return 0;
		case WM_PAINT:
			{
				POINT *pData = (POINT *) GetWindowLongPtr (hwnd, GWLP_USERDATA);
				if (pData) {
					PAINTSTRUCT paint;
					BOOL rc;
					RECT rect;
					SIZE original_port;
					GetClientRect (hwnd, &rect);
					BeginPaint (hwnd, &paint);
					SetBkColor (paint.hdc, 0);
					rc = SetMapMode (paint.hdc, MM_ANISOTROPIC);
					rc = SetWindowExtEx (paint.hdc, 2049, 600, NULL);
					SetViewportExtEx (paint.hdc, rect.right - rect.left, rect.bottom -
						rect.top, &original_port);
					RECT ViewRect;
					ViewRect.left = 0;
					ViewRect.right = 2049;
					ViewRect.top = 0;
					ViewRect.bottom = 600;
					FillRect (paint.hdc, &ViewRect, (HBRUSH) GetStockObject (BLACK_BRUSH));
					HPEN green_pen = CreatePen (PS_SOLID, 0, 0x00ff00);
					HPEN oldpen = (HPEN) SelectObject (paint.hdc, green_pen);
					Polyline (paint.hdc, pData, 2049);
					SelectObject (paint.hdc, oldpen);
					DeleteObject (green_pen);
					EndPaint (hwnd, &paint);
				}
			}
			return 0;
		case WM_DESTROY:
			{
				POINT *pData = (POINT *) GetWindowLongPtr (hwnd, GWLP_USERDATA);
				if (pData) delete[] pData;
			}
			return 0;
	}
	return DefWindowProc (hwnd, message, wParam, lParam);
}


LRESULT CALLBACK GenericButtonProc (HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam) {
	// This is my home-made button, which gets around some of the undesirable "features" of Windows' owner-draw button class
	switch (message) {
		case WM_CREATE:
			{
				CREATESTRUCT *pCreateStruct = (CREATESTRUCT *) lParam;
				SetWindowLongPtr (hwnd, GWLP_USERDATA, (LONG_PTR) (pCreateStruct->lpCreateParams));	// This will be a GenericButton object
			}
			return 0;
		case WM_ENABLE:
			InvalidateRect (hwnd, NULL, FALSE);
			return 0;
		case WM_PAINT:
			{
				GenericButton *pButton = (GenericButton *) GetWindowLongPtr (hwnd, GWLP_USERDATA);
				if (pButton) {
					PAINTSTRUCT Paint;
					BeginPaint (hwnd, &Paint);
					pButton->Draw (Paint.hdc);
					EndPaint (hwnd, &Paint);
				}
			}
			return 0;
		case WM_LBUTTONDOWN:
			{
				GenericButton *pButton = (GenericButton *) GetWindowLongPtr (hwnd, GWLP_USERDATA);
				if (pButton) {
					pButton->Press ();
					SetCapture (hwnd);
				}
			}
			return 0;
		case WM_LBUTTONUP:
			{
				ReleaseCapture ();
				GenericButton *pButton = (GenericButton *) GetWindowLongPtr (hwnd, GWLP_USERDATA);
				if (pButton) {
					pButton->Release ();
				}
			}
			return 0;
		case WM_MOUSEMOVE:
			{
				GenericButton *pButton = (GenericButton *) GetWindowLongPtr (hwnd, GWLP_USERDATA);
				if (pButton) {
					RECT Rect;
					GetClientRect (hwnd, &Rect);
					POINT Pt;
					Pt.x = short (LOWORD (lParam));
					Pt.y = short (HIWORD (lParam));
					pButton->SetFocus (PtInRect (&Rect, Pt) == TRUE);
				}
			}
			return 0;

		case BM_CLICK:
			SendMessage (hwnd, WM_LBUTTONDOWN, 0, 0);
			SendMessage (hwnd, WM_LBUTTONUP, 0, 0);
			return 0;

		case WM_DESTROY:
			{
				GenericButton *pButton = (GenericButton *) GetWindowLongPtr (hwnd, GWLP_USERDATA);
				if (pButton) delete pButton;
			}
			break;
	}
	return DefWindowProc (hwnd, message, wParam, lParam);
}


LRESULT CALLBACK DlsBoxProc (HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam) {
	switch (message) {
		case WM_CREATE:
			{
				RECT ClientRect;
				GetClientRect (hwnd, &ClientRect);
				int LineWidth = ClientRect.right - ClientRect.left;
				int LineHeight = (ClientRect.bottom - ClientRect.top) / 5;
				CreateWindowEx (0, L"static", L"", WS_CHILD | WS_VISIBLE | SS_CENTER | SS_NOPREFIX, 0, 0, LineWidth, LineHeight, hwnd, (HMENU) DLS_LINE_1, Instance, NULL);
				CreateWindowEx (0, L"static", L"", WS_CHILD | WS_VISIBLE | SS_CENTER | SS_NOPREFIX, 0, LineHeight, LineWidth, LineHeight, hwnd, (HMENU) DLS_LINE_2, Instance, NULL);
				CreateWindowEx (0, L"static", L"", WS_CHILD | WS_VISIBLE | SS_CENTER | SS_NOPREFIX, 0, 2 * LineHeight, LineWidth, LineHeight, hwnd, (HMENU) DLS_LINE_3, Instance, NULL);
				CreateWindowEx (0, L"static", L"", WS_CHILD | WS_VISIBLE | SS_CENTER | SS_NOPREFIX, 0, 3 * LineHeight, LineWidth, LineHeight, hwnd, (HMENU) DLS_LINE_4, Instance, NULL);
				CreateWindowEx (0, L"static", L"", WS_CHILD | WS_VISIBLE | SS_CENTER | SS_NOPREFIX, 0, 4 * LineHeight, LineWidth, LineHeight, hwnd, (HMENU) DLS_LINE_5, Instance, NULL);
				SendDlgItemMessage (hwnd, DLS_LINE_1, WM_SETFONT, (WPARAM) hFont11, 0);
				SendDlgItemMessage (hwnd, DLS_LINE_2, WM_SETFONT, (WPARAM) hFont11, 0);
				SendDlgItemMessage (hwnd, DLS_LINE_3, WM_SETFONT, (WPARAM) hFont11, 0);
				SendDlgItemMessage (hwnd, DLS_LINE_4, WM_SETFONT, (WPARAM) hFont11, 0);
				SendDlgItemMessage (hwnd, DLS_LINE_5, WM_SETFONT, (WPARAM) hFont11, 0);
			}
			return 0;
		case WM_CTLCOLORSTATIC:
			{
				HDC hdc = (HDC) wParam;
				SetBkColor (hdc, RGB (255, 250, 228));
				SetTextColor (hdc, RGB (0, 0, 0));
				return (LRESULT) hOffwhiteBrush;
			}
	}
	return DefWindowProc (hwnd, message, wParam, lParam);
}


LRESULT CALLBACK MotBoxProc (HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam) {
	switch (message) {
		case WM_CREATE:
			SetWindowLongPtr (hwnd, GWLP_USERDATA, NULL);
			return 0;
		case WM_PAINT:
			{
				PAINTSTRUCT ps;
				HDC hdc = BeginPaint(hwnd, &ps);
				Image *pImage = (Image *) GetWindowLongPtr (hwnd, GWLP_USERDATA);
				if (pImage) {
					Graphics graphics(hdc);
					if (graphics.GetLastStatus () == Ok) {
						RECT ClientRect;
						GetClientRect (hwnd, &ClientRect);
						// Need to specify width and height otherwise DrawImage will scale it according to the image's DPI value
						graphics.DrawImage(pImage, 0, 0, ClientRect.right - ClientRect.left, ClientRect.bottom - ClientRect.top);
					}
				}
				EndPaint(hwnd, &ps);
			}
			return 0;
		case UM_SLIDESHOW:
			{
				// Replace the old image with the new one and force a repaint
				Image *pOldImage = (Image *) GetWindowLongPtr (hwnd, GWLP_USERDATA);
				if (pOldImage) delete pOldImage;
				SetWindowLongPtr (hwnd, GWLP_USERDATA, lParam);
				if (pOldImage || lParam) InvalidateRect (hwnd, NULL, TRUE);
			}
			return 0;
		case WM_CLOSE:
			{
				Image *pOldImage = (Image *) GetWindowLongPtr (hwnd, GWLP_USERDATA);
				if (pOldImage) delete pOldImage;
				SetWindowLongPtr (hwnd, GWLP_USERDATA, NULL);
			}
			return 0;
	}
	return DefWindowProc (hwnd, message, wParam, lParam);
}


LRESULT CALLBACK InfoBoxProc (HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam) {
	switch (message) {
		case WM_CREATE:
			{
				RECT ClientRect;
				GetClientRect (hwnd, &ClientRect);
				int LineWidth = ClientRect.right - ClientRect.left;
				int LineHeight = (ClientRect.bottom - ClientRect.top) / 9;
				CreateWindowEx (0, L"static", L"", WS_CHILD | WS_VISIBLE | SS_CENTER | SS_NOPREFIX, 0, 0, LineWidth, LineHeight, hwnd, (HMENU) DATE_TIME, Instance, NULL);
				CreateWindowEx (0, L"static", L"", WS_CHILD | WS_VISIBLE | SS_CENTER | SS_NOPREFIX, 0, LineHeight, LineWidth, LineHeight, hwnd, (HMENU) SERVICE_NAME, Instance, NULL);
				CreateWindowEx (0, L"static", L"", WS_CHILD | WS_VISIBLE | SS_CENTER | SS_NOPREFIX, 0, 2 * LineHeight, LineWidth, LineHeight, hwnd, (HMENU) ENSEMBLE_ID, Instance, NULL);
				CreateWindowEx (0, L"static", L"", WS_CHILD | WS_VISIBLE | SS_CENTER | SS_NOPREFIX, 0, 3 * LineHeight, LineWidth, LineHeight, hwnd, (HMENU) SERVICE_ID, Instance, NULL);
				CreateWindowEx (0, L"static", L"", WS_CHILD | WS_VISIBLE | SS_CENTER | SS_NOPREFIX, 0, 4 * LineHeight, LineWidth, LineHeight, hwnd, (HMENU) SUBCHANNEL_ID, Instance, NULL);
				CreateWindowEx (0, L"static", L"", WS_CHILD | WS_VISIBLE | SS_CENTER | SS_NOPREFIX, 0, 5 * LineHeight, LineWidth, LineHeight, hwnd, (HMENU) SAMPLING_RATE, Instance, NULL);
				CreateWindowEx (0, L"static", L"", WS_CHILD | WS_VISIBLE | SS_CENTER | SS_NOPREFIX, 0, 6 * LineHeight, LineWidth, LineHeight, hwnd, (HMENU) BIT_RATE, Instance, NULL);
				CreateWindowEx (0, L"static", L"", WS_CHILD | WS_VISIBLE | SS_CENTER | SS_NOPREFIX, 0, 7 * LineHeight, LineWidth, LineHeight, hwnd, (HMENU) FORMAT, Instance, NULL);
				CreateWindowEx (0, L"static", L"", WS_CHILD | WS_VISIBLE | SS_CENTER | SS_NOPREFIX, 0, 8 * LineHeight, LineWidth, LineHeight, hwnd, (HMENU) UNCORRECTED_ERRORS, Instance, NULL);
				SendDlgItemMessage (hwnd, DATE_TIME, WM_SETFONT, (WPARAM) hFont11, 0);
				SendDlgItemMessage (hwnd, SERVICE_NAME, WM_SETFONT, (WPARAM) hFont11, 0);
				SendDlgItemMessage (hwnd, ENSEMBLE_ID, WM_SETFONT, (WPARAM) hFont11, 0);
				SendDlgItemMessage (hwnd, SERVICE_ID, WM_SETFONT, (WPARAM) hFont11, 0);
				SendDlgItemMessage (hwnd, SUBCHANNEL_ID, WM_SETFONT, (WPARAM) hFont11, 0);
				SendDlgItemMessage (hwnd, SAMPLING_RATE, WM_SETFONT, (WPARAM) hFont11, 0);
				SendDlgItemMessage (hwnd, BIT_RATE, WM_SETFONT, (WPARAM) hFont11, 0);
				SendDlgItemMessage (hwnd, FORMAT, WM_SETFONT, (WPARAM) hFont11, 0);
				SendDlgItemMessage (hwnd, UNCORRECTED_ERRORS, WM_SETFONT, (WPARAM) hFont11, 0);
			}
			return 0;
		case WM_CTLCOLORSTATIC:
			{
				HDC hdc = (HDC) wParam;
				SetBkColor (hdc, RGB (255, 250, 228));
				SetTextColor (hdc, RGB (0, 0, 0));
				return (LRESULT) hOffwhiteBrush;
			}
	}
	return DefWindowProc (hwnd, message, wParam, lParam);
}


int Brightness (int x, int y) {
	float fx = float (x);
	float fy = float (y);
	return int ((16 - fx * fx /400) * (16 - fy * fy / 56)) - 1;
}

void ChannelButton::Draw (HDC hDC) {
	RECT Rect;
	GetClientRect (GetDlgItem (hParent, CtlID), &Rect);
	int SavedDC = SaveDC (hDC);
	SetBkMode (hDC, TRANSPARENT);

	SIZE rsize;
	rsize.cx = Rect.right - Rect.left;
	rsize.cy = Rect.bottom - Rect.top;

	SIZE offset;
	
	HBRUSH BackgroundBrush = hUnlitBrush;

	if (Pressed && HasFocus) {
		if (CtlID == ChannelSelection) {
			for (int x = 1; x < 79; x++) {
				for (int y = 1; y < rsize.cy - 1; y++) {
					int Level = Brightness (x - 41, y - rsize.cy / 2);
					SetPixel (hDC, x, y, RGB (Level, Level, Level / 4));
				}
			}
		}
		else FillRect (hDC, &Rect, BackgroundBrush);
		for (int x = Rect.left; x < Rect.right; x++) SetPixel (hDC, x, Rect.top, RGB (0, 0, 0));
		for (int y = Rect.top; y < Rect.bottom; y++) SetPixel (hDC, Rect.left, y, RGB (0, 0, 0));
		for (int x = Rect.left + 1; x < Rect.right; x++) SetPixel (hDC, x, Rect.top + 1, RGB (80, 80, 80));
		for (int y = Rect.top + 2; y < Rect.bottom; y++) SetPixel (hDC, Rect.left + 1, y, RGB (80, 80, 80));
		for (int x = Rect.left; x < Rect.right; x++) SetPixel (hDC, x, Rect.bottom - 1, RGB (80, 80, 80));
		for (int y = Rect.top; y < Rect.bottom; y++) SetPixel (hDC, Rect.right - 1, y, RGB (80, 80, 80));
		offset.cx = 1;
		offset.cy = 1;
	}
	else {
		if (CtlID == ChannelSelection) {
			// Light it up
			for (int x = 1; x < 79; x++) {
				for (int y = 1; y < rsize.cy - 1; y++) {
					int Level = Brightness (x - 40, y - rsize.cy / 2);
					SetPixel (hDC, x, y, RGB (Level, Level, Level / 4));
				}
			}
		}
		else FillRect (hDC, &Rect, BackgroundBrush);
		for (int x = Rect.left; x < Rect.right; x++) SetPixel (hDC, x, Rect.top, RGB (220, 220, 220));
		for (int y = Rect.top; y < Rect.bottom; y++) SetPixel (hDC, Rect.left, y, RGB (220, 220, 220));
		for (int x = Rect.left; x < Rect.right; x++) SetPixel (hDC, x, Rect.bottom - 1, RGB (0, 0, 0));
		for (int y = Rect.top; y < Rect.bottom; y++) SetPixel (hDC, Rect.right - 1, y, RGB (0, 0, 0));
		for (int x = Rect.left + 1; x < Rect.right; x++) SetPixel (hDC, x, Rect.bottom - 2, RGB (80, 80, 80));
		for (int y = Rect.top; y < Rect.bottom - 2; y++) SetPixel (hDC, Rect.right - 2, y, RGB (80, 80, 80));
		offset.cx = 0;
		offset.cy = 0;
	}

	SIZE sz;
	POINT pos;
	WCHAR Caption[256];
	SelectObject (hDC, (HGDIOBJ) (hFont12));
	GetDlgItemText (hParent, CtlID, Caption, sizeof (Caption));
	GetTextExtentPoint32 (hDC, Caption, (int) wcslen (Caption), &sz);
	pos.x = (rsize.cx - sz.cx) / 2;
	pos.y = (rsize.cy - sz.cy) / 2;
	SetTextColor (hDC, RGB (0, 0, 0));
	TextOut (hDC, pos.x + offset.cx, pos.y + offset.cy, Caption, (int) wcslen (Caption));

	RestoreDC (hDC, SavedDC);
}


SegmentEntry::~SegmentEntry () {
	if (Container) {
		delete[] Container;
	}
}


void PadDecoder::FlushPad () {
	Dls.Flush ();
	MOT.Flush ();
	XPadLength = 0;
	XPadInd = 0;
	Music = false;
	Speech = false;
	CI = false;
	NextAppTy = 0;
}


void PadDecoder::DecodeFPad (unsigned char *FPad, unsigned DseLength) {
	switch (FPad[0] & 0xc0) {
		case 0:
			// F-PAD type "00"
			XPadInd = FPad[0] & 0x30;
			switch (XPadInd) {
				case 0:
				case 0x30:
					XPadLength = 0;
					break;
				case 0x10:
					XPadLength = 4;
					break;
				case 0x20:
					XPadLength = DseLength < 198 ? DseLength - 2 : 196;
					break;
			}
			break;
		case 0x80:
			// F-PAD type "10"
			switch (FPad[0] & 0x30) {
				case 0x10:
					Music = (FPad[0] & 0x04) != 0;
					Speech = (FPad[0] & 0x08) != 0;
					break;
			}
			break;
	}
	CI = (FPad[1] & 0x02) != 0;
}


void PadDecoder::DecodeXPad (unsigned char *XPad, unsigned FieldLength) {
	unsigned const LengthTable[8] = { 4, 6, 8, 12, 16, 24, 32, 48 };
	switch (XPadInd) {
		case 0x10:
			// Fixed X-PAD
			if (CI) {
				NextAppTy = XPad[0];
				DecodeSubfield (XPad + 1, NextAppTy, 3, false);
			}
			else DecodeSubfield (XPad, NextAppTy, 4, true);
			break;
		case 0x20:
			// Variable X-PAD
			if (CI) {
				unsigned AppTy[4];
				unsigned SubfieldLength[4];
				unsigned CICount = 0;
				unsigned char *pXPad;
				for (pXPad = XPad; (pXPad < XPad + 4) && (*pXPad & 0x1f); ++pXPad, ++CICount) {
					SubfieldLength[CICount] = LengthTable[*pXPad >> 5];
					AppTy[CICount] = *pXPad & 0x1f;
					if (AppTy[CICount] == 0x1f) {
						AppTy[CICount] = *(++pXPad) - 31;
					}
				}
				if (pXPad < XPad + 4) ++pXPad;	// CI list ended in NULL CI
				for (unsigned i = 0; i < CICount; i++) {
					DecodeSubfield (pXPad, AppTy[i], SubfieldLength[i], false);
					pXPad += SubfieldLength[i];
				}
				NextAppTy = CICount > 0 ? AppTy[CICount - 1] : 0;
			}
			else {
				DecodeSubfield (XPad, NextAppTy, FieldLength, true);
			}
			break;
	}
}


void PadDecoder::DecodeSubfield (unsigned char *pXPad, unsigned AppTy, unsigned Length, bool Continuation) {
	switch (AppTy) {
		case 1:
			MOT.Start (pXPad, Length, Continuation);
			break;
		case 2:
			// DLS start of data group
			if (Continuation) Dls.Continue (pXPad, Length);
			else Dls.Start (pXPad, Length);
			break;
		case 3:
			// DLS continuation of data group
			Dls.Continue (pXPad, Length);
			break;
		case 12:
		case 13:
			MOT.Continue (pXPad, Length);
			break;
	}
}


DlsDecoder::~DlsDecoder () {
	if (DlsWindow) {
		SetDlgItemText (DlsWindow, DLS_LINE_1, L"");
		SetDlgItemText (DlsWindow, DLS_LINE_2, L"");
		SetDlgItemText (DlsWindow, DLS_LINE_3, L"");
		SetDlgItemText (DlsWindow, DLS_LINE_4, L"");
		SetDlgItemText (DlsWindow, DLS_LINE_5, L"");
	}
}


void DlsDecoder::Start (unsigned char *pXPad, unsigned Length) {
	Crc = 0xffff;
	unsigned T = *pXPad & 0x80;
	FirstLast = *pXPad & 0x60;
	unsigned CFlag = *pXPad & 0x10;
	if (CFlag) {
		if ((*pXPad & 0x0f) == 1) EraseDls ();
	}
	else SegmentLength = *pXPad & 0x0f;
	ByteCount = 1;
	unsigned short Data = (unsigned short) (*pXPad) << 8;
	for (int Bit = 0; Bit < 8; Bit++) {
		bool Feedback = ((Crc ^ Data) & 0x8000) != 0;
		Crc <<=1;
		Data <<= 1;
		if (Feedback) Crc ^= 0x1021;
	}
	if (Length > 0) Continue (pXPad + 1, Length - 1);
}


bool ConvertToWide (WCHAR *Destin, char *Source, unsigned Charset) {
	// Convert to Unicode based on the DLS Charset field.
	// EBU Latin 2 and EBU Latin 3 are no longer part of the DAB standard, but no harm leaving them in here.
	switch (Charset) {
		case 0:
			for (int i = 0; i < 132; i++) {
				Destin[i] = EBULatin1[(unsigned char) Source[i]];
			}
			break;
		case 1:
			for (int i = 0; i < 132; i++) {
				Destin[i] = EBULatin2[(unsigned char) Source[i]];
			}
			break;
		case 2:
			for (int i = 0; i < 132; i++) {
				Destin[i] = EBULatin3[(unsigned char) Source[i]];
			}
			break;
		case 3:
			return MultiByteToWideChar (28592, 0, Source, 132, Destin, 132) > 0;
		case 6:
			for (int i = 0; i < 66; i++) {
				Destin[i] = (((WCHAR) Source[i * 2]) << 8) | (unsigned char) Source[i * 2 + 1];
			}
			Destin[66] = 0;
			break;
		case 15:
			return MultiByteToWideChar (65001, 0, Source, 132, Destin, 132) > 0;
	}
	return true;
}


void DlsDecoder::Continue (unsigned char *pXPad, unsigned Length) {
	while (Length-- > 0) {
		if (ByteCount == 0) break;
		unsigned short Data = (unsigned short) (*pXPad) << 8;
		for (int Bit = 0; Bit < 8; Bit++) {
			bool Feedback = ((Crc ^ Data) & 0x8000) != 0;
			Crc <<=1;
			Data <<= 1;
			if (Feedback) Crc ^= 0x1021;
		}
		if (ByteCount == 1) {
			if (FirstLast & 0x40) {
				Charset = (*pXPad & 0xf0) >> 4;
				SegNum = 0;
				memset (ValidSegment, 0, sizeof ValidSegment);
				memset (SegLen, 0, sizeof SegLen);
			}
			else SegNum = (*pXPad & 0x70) >> 4;
		}
		else if (ByteCount < SegmentLength + 3) {
			DlsText[SegNum][ByteCount - 2] = *pXPad;
		}
		else if (ByteCount == SegmentLength + 4) {
			// Check for valid CRC and display it
			if (Crc == 0x1d0f) {
				ValidSegment[SegNum] = true;
				SegLen[SegNum] = SegmentLength + 1;
				if (FirstLast & 0x20) {
					char LinearString[132] = { 0 };
					WCHAR Display[5][33] = { 0 };
					unsigned Row = 0;
					unsigned Col = 0;
					unsigned LastSpace = 0;
					unsigned WordBreak = 0;
					bool Headline = false;
					unsigned Seg;
					char *pLinear = LinearString;
					for (Seg = 0; Seg <= SegNum; Seg++) {
						if (!ValidSegment[Seg]) break;
						for (unsigned Letter = 0; Letter < SegLen[Seg]; Letter++) {
							*(pLinear++) = DlsText[Seg][Letter];
						}
					}
					WCHAR WideString[132] = { 0 };
					if (Seg > SegNum && ConvertToWide (WideString, LinearString, Charset)) {
						for (unsigned Letter = 0; WideString[Letter] && Letter < 128; Letter++) {
							if (WideString[Letter] == L' ') LastSpace = Col;
							switch (WideString[Letter]) {
								case 0x000a:
									if (Row < 4) {
										Row++;
										Col = 0;
									}
									else if (Col < 32) Display[Row][Col++] = L' ';
									break;
								case 0x000b:
									Headline = true;
									break;
								case 0x001f:
									WordBreak = Col;
									break;
								default:
									if (Col < 32) Display[Row][Col++] = WideString[Letter];
									else if (Row < 4) {
										if (WideString[Letter] == L' ') {
											Display[Row][Col] = 0;
											++Row;
											Col = 0;
											LastSpace = 0;
										}
										else if (WordBreak && Row < 4) {
											Col = 0;
											++Row;
											for (unsigned Wrap = WordBreak; Wrap < 32; Wrap++) {
												Display[Row][Col++] = Display[Row - 1][Wrap];
											}
											Display[Row - 1][WordBreak] = '-';
											Display[Row - 1][WordBreak + 1] = 0;
											Display[Row][Col++] = WideString[Letter];
											WordBreak = 0;
											LastSpace = 0;
										}
										else if (LastSpace) {
											Display[Row][LastSpace] = 0;
											Col = 0;
											++Row;
											for (unsigned Wrap = LastSpace + 1; Wrap < 32; Wrap++) {
												Display[Row][Col++] = Display[Row - 1][Wrap];
											}
											Display[Row][Col++] = WideString[Letter];
											LastSpace = 0;
										}
										else if (Row < 4) {
											++Row;
											Col = 0;
											Display[Row][Col++] = WideString[Letter];
										}
									}
							}
						}
						Display[Row][Col] = 0;
						while (Row < 4) Display[++Row][0] = 0;
						if (DlsWindow) {
							SetDlgItemTextW (DlsWindow, DLS_LINE_1, Display[0]);
							SetDlgItemTextW (DlsWindow, DLS_LINE_2, Display[1]);
							SetDlgItemTextW (DlsWindow, DLS_LINE_3, Display[2]);
							SetDlgItemTextW (DlsWindow, DLS_LINE_4, Display[3]);
							SetDlgItemTextW (DlsWindow, DLS_LINE_5, Display[4]);
						}
					}
				}
			}
			break;
		}
		++ByteCount;
		++pXPad;
	}
}


void DlsDecoder::EraseDls () {
	if (DlsWindow) {
		SetDlgItemText (DlsWindow, DLS_LINE_1, L"");
		SetDlgItemText (DlsWindow, DLS_LINE_2, L"");
		SetDlgItemText (DlsWindow, DLS_LINE_3, L"");
		SetDlgItemText (DlsWindow, DLS_LINE_4, L"");
		SetDlgItemText (DlsWindow, DLS_LINE_5, L"");
	}
}


SlideDecoder::~SlideDecoder () {
	if (MotWindow) SendMessage (MotWindow, UM_SLIDESHOW, 0, 0);
	if (CurrentDataGroup) delete[] CurrentDataGroup;
}


void SlideDecoder::Start (unsigned char *pXPad, unsigned Length, bool Continuation) {
	if (!Continuation) DataGroupLengthIndicatorByteCount = 0;
	for (unsigned i = 0; i < Length && DataGroupLengthIndicatorByteCount < 4; i++, DataGroupLengthIndicatorByteCount++) {
		DataGroupLengthIndicator[DataGroupLengthIndicatorByteCount] = pXPad[i];
	}
	if (DataGroupLengthIndicatorByteCount == 4) {
		unsigned short Crc = 0xffff;
		for (unsigned i = 0; i < 4; i++) {
			unsigned short Data = (unsigned short) (DataGroupLengthIndicator[i]) << 8;
			for (int Bit = 0; Bit < 8; Bit++) {
				bool Feedback = ((Crc ^ Data) & 0x8000) != 0;
				Crc <<=1;
				Data <<= 1;
				if (Feedback) Crc ^= 0x1021;
			}
		}
		DataGroupLength = Crc == 0x1d0f ? (DataGroupLengthIndicator[0] << 8) | DataGroupLengthIndicator[1] : 0;
		if (DataGroupLength) {
			if (CurrentDataGroup) delete[] CurrentDataGroup;
			CurrentDataGroup = new UCHAR[DataGroupLength];
			CurrentDataGroupIndex = 0;
		}
	}
}


void SlideDecoder::Continue (unsigned char *pXPad, unsigned Length) {
	if (CurrentDataGroup) {
		for (unsigned i = 0; i < Length && CurrentDataGroupIndex < DataGroupLength; i++, CurrentDataGroupIndex++) {
			CurrentDataGroup[CurrentDataGroupIndex] = pXPad[i];
		}
		if (CurrentDataGroupIndex == DataGroupLength) {
			DecodeDataGroup (CurrentDataGroup, DataGroupLength);
			// At this point either CurrentDataGroup has been deleted or linked into an MotObject
			CurrentDataGroup = NULL;
		}
	}
}


void SlideDecoder::DecodeDataGroup (unsigned char *DataGroup, unsigned Length) {
	unsigned short Crc;
	if (*DataGroup & 0x40) {	// CRCFlag
		Crc = 0xffff;
		for (unsigned i = 0; i < Length; i++) {
			unsigned short Data = (unsigned short) (DataGroup[i]) << 8;
			for (int Bit = 0; Bit < 8; Bit++) {
				bool Feedback = ((Crc ^ Data) & 0x8000) != 0;
				Crc <<=1;
				Data <<= 1;
				if (Feedback) Crc ^= 0x1021;
			}
		}
	}
	else Crc = 0x1d0f;	// No CRC transmitted so pretend it was correct
	if (Crc == 0x1d0f) {
		// Analyse Data Group Header
		UCHAR *pHeader = DataGroup;
		UCHAR ExtensionFlag = *pHeader & 0x80;
		UCHAR CRCFlag = *pHeader & 0x40;
		UCHAR SegmentFlag = *pHeader & 0x20;
		UCHAR UserAccessFlag = *pHeader & 0x10;
		UCHAR DataGroupType = *pHeader & 0x0f;
		UCHAR ContinuityIndex = (*(++pHeader) & 0xf0) >> 4;
		UCHAR RepetitionIndex = *pHeader & 0x0f;
		++pHeader;
		USHORT Extension;
		if (ExtensionFlag) {
			Extension = ((USHORT)(pHeader[0]) << 8) | pHeader[1];
			pHeader += 2;
		}
		else Extension = 0;
		// Analyse Session Header
		UCHAR LastSegment;
		USHORT SegmentNumber;
		if (SegmentFlag) {
			LastSegment = pHeader[0] & 0x80;
			SegmentNumber = ((USHORT)(pHeader[0] & 0x7f) << 8) | pHeader[1];
			pHeader += 2;
		}
		else {
			LastSegment = 0x80;
			SegmentNumber = 0;
		}
		USHORT TransportId;
		if (UserAccessFlag) {
			UCHAR TransportIdFlag = *pHeader & 0x10;
			UCHAR LengthIndicator = *pHeader & 0x0f;
			++pHeader;
			if (TransportIdFlag) {
				TransportId = ((USHORT)(pHeader[0]) << 8) | pHeader[1];
			}
			pHeader += LengthIndicator;
		}
		else TransportId = 0;
		// Analyse Segmentation Header
		UCHAR RepetitionCount = pHeader[0] >> 5;
		USHORT SegmentSize = ((USHORT)(pHeader[0] & 0x1f) << 8) | pHeader[1];
		pHeader += 2;
		if (DataGroupType == 3) {
			if (TransportId != CurrentTransportId) {
				// Start of a new object so get rid of the old one
				if (HeaderSegment.Container) {
					delete[] HeaderSegment.Container;
					HeaderSegment.Container = NULL;
					HeaderSegment.Size = 0;
					HeaderSegment.Offset = 0;
				}
				for (unsigned i = 0; i < 32768; i++) {
					if (BodySegment[i].Container) {
						delete[] BodySegment[i].Container;
						BodySegment[i].Container = NULL;
						BodySegment[i].Size = 0;
						BodySegment[i].Offset = 0;
					}
				}
				SegmentCount = 0;
				CurrentTransportId = TransportId;
			}
			// Analyse header core
			unsigned BodySize = ((unsigned)pHeader[0] << 20) | ((unsigned)pHeader[1] << 12) | ((unsigned)pHeader[2] << 4) | ((unsigned)pHeader[3] >> 4);
			unsigned HeaderSize = ((unsigned)(pHeader[3] & 0x0f) << 9) | ((unsigned)pHeader[4] << 1) | ((unsigned)pHeader[5] >> 7);
			unsigned ContentType = (unsigned)(pHeader[5] & 0x7e) >> 1;
			unsigned ContentSubType = ((unsigned)(pHeader[5] & 0x01) << 8) | ((unsigned)pHeader[6]);
			if (HeaderSegment.Container == NULL && !(ContentType == 5 && ContentSubType == 0)) {
				// Don't create a new header object if this is a header update frame
				HeaderSegment.Container = DataGroup;
				HeaderSegment.Offset = pHeader - DataGroup;
				HeaderSegment.Size = SegmentSize;
			}
			else {
				// It is a repetition of the header
				delete[] DataGroup;
			}
		}
		else if (DataGroupType == 4) {
			if (CurrentTransportId == -1) CurrentTransportId = TransportId;	// Just starting
			if (TransportId == CurrentTransportId && BodySegment[SegmentNumber].Container == NULL) {
				BodySegment[SegmentNumber].Container = DataGroup;
				BodySegment[SegmentNumber].Offset = pHeader - DataGroup;
				BodySegment[SegmentNumber].Size = SegmentSize;
				if (LastSegment && DataGroupType == 4) SegmentCount = SegmentNumber + 1;
				if (HeaderSegment.Container && SegmentCount > 0) {
					bool FullObject = true;
					for (unsigned i = 0; i < SegmentCount; i++) {
						if (BodySegment[i].Container == NULL) {
							FullObject = false;
							break;
						}
					}
					if (FullObject) ProcessObject ();
				}
			}
			else {
				// It is a repetition of the body segment or a different transport ID.
				delete[] DataGroup;
			}
		}
		else {
			delete[] DataGroup;
		}
	}
	else {
		delete[] DataGroup;
	}
}


void SlideDecoder::ProcessObject () {
	if (MotWindow) {
		// Analyse header core
		UCHAR *pHeader = HeaderSegment.Container + HeaderSegment.Offset;
		unsigned BodySize = ((unsigned)pHeader[0] << 20) | ((unsigned)pHeader[1] << 12) | ((unsigned)pHeader[2] << 4) | ((unsigned)pHeader[3] >> 4);
		unsigned HeaderSize = ((unsigned)(pHeader[3] & 0x0f) << 9) | ((unsigned)pHeader[4] << 1) | ((unsigned)pHeader[5] >> 7);
		unsigned ContentType = (unsigned)(pHeader[5] & 0x7e) >> 1;
		unsigned ContentSubType = ((unsigned)(pHeader[5] & 0x01) << 8) | ((unsigned)pHeader[6]);
		pHeader += 7;
		//Analyse header extension
		int ExtensionSize = HeaderSegment.Size - 7;
		UCHAR *pExtensionBase = pHeader;
		UCHAR CompressionId = 0;
		char *Name = NULL;
		while (pHeader - pExtensionBase < ExtensionSize) {
			UCHAR PLI = pHeader[0] & 0xc0;
			UCHAR ParamId = pHeader[0] & 0x3f;
			UCHAR *pDataField;
			unsigned DataFieldLength;
			switch (PLI) {
				case 0:
					pDataField = NULL;
					DataFieldLength = 0;
					pHeader += 1;
					break;
				case 0x40:
					pDataField = pHeader + 1;
					DataFieldLength = 1;
					pHeader += 2;
					break;
				case 0x80:
					pDataField = pHeader + 1;
					DataFieldLength = 4;
					pHeader += 5;
					break;
				case 0xc0:
					if (pHeader[1] & 0x80) {
						DataFieldLength = ((unsigned)(pHeader[1] & 0x7f) << 8) | pHeader[2];
						pDataField = pHeader + 3;
						pHeader += DataFieldLength + 3;
					}
					else {
						DataFieldLength = pHeader[1] & 0x7f;
						pDataField = pHeader + 2;
						pHeader += DataFieldLength + 2;
					}
					break;
			}
			switch (ParamId) {
				case 0x03:
					// Start Validity
					break;
				case 0x05:
					// Trigger Time
					break;
				case 0x06:
					// Version Number
					break;
				case 0x11:
					// Compression type
					CompressionId = *pDataField;
					break;
				case 0x0c:
					// Component name
					Name = new char[DataFieldLength];
					memcpy (Name, pDataField + 1, DataFieldLength - 1);
					Name[DataFieldLength - 1] = 0;
					break;
			}
		}
		switch (ContentType) {
			case 0:
				switch (ContentSubType) {
					case 0:
						{
							HGLOBAL hMem = GlobalAlloc (GMEM_MOVEABLE, BodySize);
							if (hMem) {
								UCHAR *pMem = (UCHAR *) GlobalLock (hMem);
								UCHAR *pBase = pMem;
								for (unsigned i = 0; i < SegmentCount; i++) {
									memcpy (pMem, BodySegment[i].Container + BodySegment[i].Offset, BodySegment[i].Size);
									pMem += BodySegment[i].Size;
								}
								size_t BytesLoaded = pMem - pBase;
								assert (BytesLoaded == BodySize);
								GlobalUnlock (hMem);
								LPSTREAM hStream;
								if (CreateStreamOnHGlobal (hMem, TRUE, &hStream) == S_OK) {
									Image *pImage = new Image (hStream);
									if (pImage) {
										Status LastStatus = pImage->GetLastStatus ();
										if (LastStatus == Ok) {
											SendMessage (MotWindow, UM_SLIDESHOW, 0, (LPARAM) pImage);
										}
										else {
											delete pImage;
										}
									}
									hStream->Release ();
								}
							}
						}
						break;
				}
				break;
			case 2:
				// Image
				switch (ContentSubType) {
					case 1:
						// JFIF (JPEG)
					case 3:
						// PNG - Need to decode APNG
						{
							HGLOBAL hMem = GlobalAlloc (GMEM_MOVEABLE, BodySize);
							if (hMem) {
								UCHAR *pMem = (UCHAR *) GlobalLock (hMem);
								UCHAR *pBase = pMem;
								for (unsigned i = 0; i < SegmentCount; i++) {
									memcpy (pMem, BodySegment[i].Container + BodySegment[i].Offset, BodySegment[i].Size);
									pMem += BodySegment[i].Size;
								}
								size_t BytesLoaded = pMem - pBase;
								assert (BytesLoaded == BodySize);
								GlobalUnlock (hMem);
								LPSTREAM hStream;
								if (CreateStreamOnHGlobal (hMem, TRUE, &hStream) == S_OK) {
									Image *pImage = new Image (hStream);
									if (pImage) {
										Status LastStatus = pImage->GetLastStatus ();
										if (LastStatus == Ok) {
											SendMessage (MotWindow, UM_SLIDESHOW, 0, (LPARAM) pImage);
										}
										else {
											delete pImage;
										}
									}
									hStream->Release ();
								}
							}
						}
						break;
				}
				break;
		}
		if (Name) delete[] Name;
	}
}


void SlideDecoder::Flush () {
	if (MotWindow) SendMessage (MotWindow, UM_SLIDESHOW, 0, 0);
	if (CurrentDataGroup) {
		delete[] CurrentDataGroup;
		CurrentDataGroup = NULL;
	}
	if (HeaderSegment.Container) {
		delete HeaderSegment.Container;
		HeaderSegment.Container = NULL;
		HeaderSegment.Size = 0;
		HeaderSegment.Offset = 0;
	}
	for (unsigned i = 0; i < 32768; i++) {
		if (BodySegment[i].Container) {
			delete BodySegment[i].Container;
			BodySegment[i].Container = NULL;
			BodySegment[i].Size = 0;
			BodySegment[i].Offset = 0;
		}
	}
	DataGroupLength = 0;
	CurrentDataGroupIndex = 0;
	SegmentCount = 0;
}


void PadThread (void *param) {
	HANDLE hDev = (HANDLE) param;	// This is a duplicate of the handle used in the main thread
	PadDecoder Decoder (hwndMain);
	// Queue four PAD buffers
	PAD PadBuffer[4];
	OVERLAPPED Overlap[4] = { 0 };
	for (unsigned i = 0; i < 4; i++) {
		Overlap[i].hEvent = CreateEvent (NULL, TRUE, TRUE, NULL);
		if (DabRadcapGetPAD (hDev, &PadBuffer[i], &Overlap[i]) || GetLastError () != ERROR_IO_PENDING) {
			SetEvent (Overlap[i].hEvent);
		}
	}
	while (1) {
		DWORD BytesRead;
		for (int i = 0; i < 4; i++) {
			HANDLE WaitList[2] = { StopPad, Overlap[i].hEvent };
			DWORD WaitResult = WaitForMultipleObjects (2, WaitList, FALSE, INFINITE);
			switch (WaitResult) {
				case WAIT_OBJECT_0:
					// StopPad was set, so wait for all the buffers to come back from the driver and exit
					{
						HANDLE EventList[4] = { Overlap[0].hEvent, Overlap[1].hEvent, Overlap[2].hEvent, Overlap[3].hEvent };
						WaitForMultipleObjects (4, EventList, TRUE, INFINITE);
						CloseHandle (EventList[0]);
						CloseHandle (EventList[1]);
						CloseHandle (EventList[2]);
						CloseHandle (EventList[3]);
						CloseHandle (hDev);
					}
					return;
				case WAIT_OBJECT_0 + 1:
					if (GetOverlappedResult (hDev, &Overlap[i], &BytesRead, FALSE) && BytesRead >= 2) {
						Decoder.DecodeFPad (PadBuffer[i].FPad, BytesRead);
						Decoder.DecodeXPad (PadBuffer[i].XPad, BytesRead - 2);
					}
					else Sleep (100);	// Driver dumped the buffer (maybe loss of signal) so try again shortly
					if (DabRadcapGetPAD (hDev, &PadBuffer[i], &Overlap[i]) || GetLastError () != ERROR_IO_PENDING) {
						SetEvent (Overlap[i].hEvent);
					}
					break;
			}
		}
	}
}


ChannelButton::ChannelButton (HWND ParentWindow, DWORD Identifier, HANDLE hDev, unsigned WaveNum) : GenericButton (ParentWindow, Identifier), hDevice (hDev), WaveId (WaveNum) {
}

ChannelButton::~ChannelButton () {
	if (hDevice != INVALID_HANDLE_VALUE) {
		CloseHandle (hDevice);
	}
}

void ChannelButton::ActivatePad () {
	// Pass the PAD thread a duplicate handle so we can close ours possibly before the thread has exited
	if (hDevice != INVALID_HANDLE_VALUE) {
		HANDLE hDuplicate;
		if (DuplicateHandle (GetCurrentProcess (), hDevice, GetCurrentProcess (), &hDuplicate, 0, FALSE, DUPLICATE_SAME_ACCESS)) {
			if (_beginthread (PadThread, 0, (void *) hDuplicate) == -1L) CloseHandle (hDuplicate);	// Otherwise the thread closes hDuplicate when it exits
		}
	}
}


BOOL ChannelButton::GetDateTime (DABDateTimeInfo *pDateTime) {
	return pDateTime && hDevice != INVALID_HANDLE_VALUE ? DabRadcapGetDateTime (hDevice, pDateTime) : FALSE;
}

BOOL ChannelButton::GetFormat (DABFormat *pFormat) {
	return pFormat && hDevice != INVALID_HANDLE_VALUE ? DabRadcapGetFormat (hDevice, pFormat) : FALSE;
}

BOOL ChannelButton::GetEnsembleInfoW (EnsembleIdentificationW *pInfo) {
	return hDevice != INVALID_HANDLE_VALUE ? DabRadcapGetEnsembleInformationW (hDevice, pInfo) : FALSE;
}

BOOL ChannelButton::GetServiceInfoW (ServiceInfoW *pInfo) {
	return hDevice != INVALID_HANDLE_VALUE ? DabRadcapGetServiceInformationW (hDevice, pInfo) : FALSE;
}

BOOL ChannelButton::GetServiceStatus (ServiceStatus *pStatus) {
	return hDevice != INVALID_HANDLE_VALUE ? DabRadcapGetServiceStatus (hDevice, pStatus) : FALSE;
}

UINT ChannelButton::GetUncorrectedErrorCount () {
	return hDevice != INVALID_HANDLE_VALUE ? DabRadcapGetUncorrectedErrorCount (hDevice) : 0;
}
